"""AMQPStorm Connection.Channel0."""

import logging
import platform

from pamqp import specification
from pamqp.heartbeat import Heartbeat

from amqpstorm import __version__
from amqpstorm.base import LOCALE
from amqpstorm.base import MAX_CHANNELS
from amqpstorm.base import MAX_FRAME_SIZE
from amqpstorm.base import Stateful
from amqpstorm.compatibility import try_utf8_decode
from amqpstorm.exception import AMQPConnectionError

LOGGER = logging.getLogger(__name__)


class Channel0(object):
    """Internal Channel0 handler."""

    def __init__(self, connection, client_properties=None):
        super(Channel0, self).__init__()
        self.is_blocked = False
        self.max_allowed_channels = MAX_CHANNELS
        self.max_frame_size = MAX_FRAME_SIZE
        self.server_properties = {}
        self._connection = connection
        self._heartbeat = connection.parameters['heartbeat']
        self._parameters = connection.parameters
        self._override_client_properties = client_properties

    def on_frame(self, frame_in):
        """Handle frames sent to Channel0.

        :param frame_in: Amqp frame.
        :return:
        """
        LOGGER.debug('Frame Received: %s', frame_in.name)
        if frame_in.name == 'Heartbeat':
            return
        elif frame_in.name == 'Connection.Close':
            self._close_connection(frame_in)
        elif frame_in.name == 'Connection.CloseOk':
            self._close_connection_ok()
        elif frame_in.name == 'Connection.Blocked':
            self._blocked_connection(frame_in)
        elif frame_in.name == 'Connection.Unblocked':
            self._unblocked_connection()
        elif frame_in.name == 'Connection.OpenOk':
            self._set_connection_state(Stateful.OPEN)
        elif frame_in.name == 'Connection.Start':
            self.server_properties = frame_in.server_properties
            self._send_start_ok(frame_in)
        elif frame_in.name == 'Connection.Tune':
            self._send_tune_ok(frame_in)
            self._send_open_connection()
        else:
            LOGGER.error('[Channel0] Unhandled Frame: %s', frame_in.name)

    def send_close_connection(self):
        """Send Connection Close frame.

        :return:
        """
        self._write_frame(specification.Connection.Close())

    def send_heartbeat(self):
        """Send Heartbeat frame.

        :return:
        """
        if not self._connection.is_open:
            return
        self._write_frame(Heartbeat())

    def _close_connection(self, frame_in):
        """Connection Close.

        :param specification.Connection.Close frame_in: Amqp frame.
        :return:
        """
        self._set_connection_state(Stateful.CLOSED)
        if frame_in.reply_code != 200:
            reply_text = try_utf8_decode(frame_in.reply_text)
            message = (
                'Connection was closed by remote server: %s' % reply_text
            )
            exception = AMQPConnectionError(message,
                                            reply_code=frame_in.reply_code)
            self._connection.exceptions.append(exception)

    def _close_connection_ok(self):
        """Connection CloseOk frame received.

        :return:
        """
        self._set_connection_state(Stateful.CLOSED)

    def _blocked_connection(self, frame_in):
        """Connection is Blocked.

        :param frame_in:
        :return:
        """
        self.is_blocked = True
        LOGGER.warning(
            'Connection is blocked by remote server: %s',
            try_utf8_decode(frame_in.reason)
        )

    def _negotiate(self, server_value, client_value):
        """Negotiate the highest supported value. Fall back on the
        client side value if zero.

        :param int server_value: Server Side value
        :param int client_value: Client Side value

        :rtype: int
        :return:
        """
        return min(server_value, client_value) or client_value

    def _unblocked_connection(self):
        """Connection is Unblocked.

        :return:
        """
        self.is_blocked = False
        LOGGER.info('Connection is no longer blocked by remote server')

    def _plain_credentials(self):
        """AMQP Plain Credentials.

        :rtype: str
        """
        return '\0%s\0%s' % (self._parameters['username'],
                             self._parameters['password'])

    def _send_start_ok(self, frame_in):
        """Send Start OK frame.

        :param specification.Connection.Start frame_in: Amqp frame.
        :return:
        """
        mechanisms = try_utf8_decode(frame_in.mechanisms)
        if 'EXTERNAL' in mechanisms:
            mechanism = 'EXTERNAL'
            credentials = '\0\0'
        elif 'PLAIN' in mechanisms:
            mechanism = 'PLAIN'
            credentials = self._plain_credentials()
        else:
            exception = AMQPConnectionError(
                'Unsupported Security Mechanism(s): %s' %
                frame_in.mechanisms
            )
            self._connection.exceptions.append(exception)
            return
        start_ok_frame = specification.Connection.StartOk(
            mechanism=mechanism,
            client_properties=self._client_properties(),
            response=credentials,
            locale=LOCALE
        )
        self._write_frame(start_ok_frame)

    def _send_tune_ok(self, frame_in):
        """Send Tune OK frame.

        :param specification.Connection.Tune frame_in: Tune frame.

        :return:
        """
        self.max_allowed_channels = self._negotiate(frame_in.channel_max,
                                                    MAX_CHANNELS)
        self.max_frame_size = self._negotiate(frame_in.frame_max,
                                              MAX_FRAME_SIZE)
        LOGGER.debug(
            'Negotiated max frame size %d, max channels %d',
            self.max_frame_size, self.max_allowed_channels
        )

        tune_ok_frame = specification.Connection.TuneOk(
            channel_max=self.max_allowed_channels,
            frame_max=self.max_frame_size,
            heartbeat=self._heartbeat)
        self._write_frame(tune_ok_frame)

    def _send_open_connection(self):
        """Send Open Connection frame.

        :return:
        """
        open_frame = specification.Connection.Open(
            virtual_host=self._parameters['virtual_host']
        )
        self._write_frame(open_frame)

    def _set_connection_state(self, state):
        """Set Connection state.

        :param state:
        :return:
        """
        self._connection.set_state(state)

    def _write_frame(self, frame_out):
        """Write a pamqp frame from Channel0.

        :param frame_out: Amqp frame.
        :return:
        """
        self._connection.write_frame(0, frame_out)
        LOGGER.debug('Frame Sent: %s', frame_out.name)

    def _client_properties(self):
        """AMQPStorm Client Properties.

        :rtype: dict
        """
        client_properties = {
            'product': 'AMQPStorm',
            'platform': 'Python %s (%s)' % (platform.python_version(),
                                            platform.python_implementation()),
            'capabilities': {
                'basic.nack': True,
                'connection.blocked': True,
                'publisher_confirms': True,
                'consumer_cancel_notify': True,
                'authentication_failure_close': True,
            },
            'information': 'See https://github.com/eandersson/amqpstorm',
            'version': __version__
        }
        if self._override_client_properties:
            client_properties.update(self._override_client_properties)
        return client_properties



import platform

from pamqp.heartbeat import Heartbeat
from pamqp.specification import Connection

import amqpstorm
from amqpstorm import AMQPConnectionError
from amqpstorm.base import MAX_CHANNELS
from amqpstorm.base import MAX_FRAME_SIZE
from amqpstorm.channel0 import Channel0
from amqpstorm.tests.utility import FakeConnection
from amqpstorm.tests.utility import TestFramework


class Channel0Tests(TestFramework):
    def test_channel0_client_properties(self):
        channel = Channel0(FakeConnection())
        result = channel._client_properties()

        information = 'See https://github.com/eandersson/amqpstorm'
        python_version = 'Python %s (%s)' % (platform.python_version(),
                                             platform.python_implementation())

        self.assertIsInstance(result, dict)
        self.assertTrue(result['capabilities']['authentication_failure_close'])
        self.assertTrue(result['capabilities']['consumer_cancel_notify'])
        self.assertTrue(result['capabilities']['publisher_confirms'])
        self.assertTrue(result['capabilities']['connection.blocked'])
        self.assertTrue(result['capabilities']['basic.nack'])
        self.assertEqual(result['information'], information)
        self.assertEqual(result['platform'], python_version)

    def test_channel0_credentials(self):
        connection = FakeConnection()
        connection.parameters['username'] = 'guest'
        connection.parameters['password'] = 'password'
        channel = Channel0(connection)
        credentials = channel._plain_credentials()

        self.assertEqual(credentials, '\0guest\0password')

    def test_channel0_close_connection(self):
        connection = FakeConnection()
        connection.set_state(connection.OPEN)
        channel = Channel0(connection)

        self.assertTrue(connection.is_open)

        channel._close_connection(
            Connection.Close(reply_text=b'',
                             reply_code=200)
        )

        self.assertFalse(connection.exceptions)
        self.assertTrue(connection.is_closed)

    def test_channel0_forcefully_closed_connection(self):
        connection = amqpstorm.Connection('localhost', 'guest', 'guest',
                                          lazy=True)
        connection.set_state(connection.OPEN)
        channel = Channel0(connection)
        channel._close_connection(
            Connection.Close(reply_text=b'',
                             reply_code=500)
        )

        self.assertTrue(connection.is_closed)
        self.assertRaises(AMQPConnectionError, connection.check_for_errors)

    def test_channel0_send_start_ok_plain(self):
        connection = FakeConnection()
        connection.parameters['username'] = 'guest'
        connection.parameters['password'] = 'password'
        channel = Channel0(connection)
        channel._send_start_ok(Connection.Start(mechanisms=b'PLAIN'))

        self.assertTrue(connection.frames_out)

        channel_id, frame_out = connection.frames_out.pop()

        self.assertEqual(channel_id, 0)
        self.assertIsInstance(frame_out, Connection.StartOk)
        self.assertNotEqual(frame_out.locale, '')
        self.assertIsNotNone(frame_out.locale)

    def test_channel0_send_start_ok_external(self):
        connection = FakeConnection()
        channel = Channel0(connection)
        channel._send_start_ok(Connection.Start(mechanisms=b'EXTERNAL'))

        self.assertTrue(connection.frames_out)

        channel_id, frame_out = connection.frames_out.pop()

        self.assertEqual(channel_id, 0)
        self.assertIsInstance(frame_out, Connection.StartOk)
        self.assertNotEqual(frame_out.locale, '')
        self.assertIsNotNone(frame_out.locale)

    def test_channel0_send_tune_ok(self):
        connection = FakeConnection()
        channel = Channel0(connection)
        channel._send_tune_ok(Connection.Tune())

        self.assertTrue(connection.frames_out)

        channel_id, frame_out = connection.frames_out.pop()

        self.assertEqual(channel_id, 0)
        self.assertIsInstance(frame_out, Connection.TuneOk)

        self.assertEqual(frame_out.channel_max, MAX_CHANNELS)
        self.assertEqual(frame_out.frame_max, MAX_FRAME_SIZE)

    def test_channel0_send_tune_ok_negotiate(self):
        channel = Channel0(FakeConnection())
        channel._send_tune_ok(Connection.Tune(frame_max=MAX_FRAME_SIZE,
                                              channel_max=MAX_CHANNELS))

        self.assertEqual(channel.max_frame_size, MAX_FRAME_SIZE)
        self.assertEqual(channel.max_allowed_channels, MAX_CHANNELS)

    def test_channel0_send_tune_ok_negotiate_use_max(self):
        """Test to make sure that we use the highest acceptable value when
        the server returns zero.
        """
        channel = Channel0(FakeConnection())
        channel._send_tune_ok(Connection.Tune())

        self.assertEqual(channel.max_frame_size, MAX_FRAME_SIZE)
        self.assertEqual(channel.max_allowed_channels, MAX_CHANNELS)

    def test_channel0_send_tune_ok_negotiate_use_server(self):
        """Test to make sure that we use the highest acceptable value from
        the servers perspective.
        """
        channel = Channel0(FakeConnection())
        channel._send_tune_ok(Connection.Tune(
            frame_max=16384,
            channel_max=200
        ))

        self.assertEqual(channel.max_frame_size, 16384)
        self.assertEqual(channel.max_allowed_channels, 200)

    def test_channel0_send_heartbeat(self):
        connection = FakeConnection()
        channel = Channel0(connection)
        channel.send_heartbeat()

        self.assertTrue(connection.frames_out)

        channel_id, frame_out = connection.frames_out.pop()

        self.assertEqual(channel_id, 0)
        self.assertIsInstance(frame_out, Heartbeat)

    def test_channel0_do_not_send_heartbeat_when_connection_closed(self):
        connection = FakeConnection(state=FakeConnection.CLOSED)
        channel = Channel0(connection)
        channel.send_heartbeat()

        self.assertFalse(connection.frames_out)

    def test_channel0_send_close_connection(self):
        connection = FakeConnection()
        channel = Channel0(connection)
        channel.send_close_connection()

        self.assertTrue(connection.frames_out)

        channel_id, frame_out = connection.frames_out.pop()

        self.assertEqual(channel_id, 0)
        self.assertIsInstance(frame_out, Connection.Close)

    def test_channel0_invalid_authentication_mechanism(self):
        connection = amqpstorm.Connection('localhost', 'guest', 'guest',
                                          lazy=True)
        channel = Channel0(connection)
        channel._send_start_ok(
            Connection.Start(mechanisms='CRAM-MD5 SCRAM-SHA-1 SCRAM-SHA-256')
        )

        self.assertRaises(AMQPConnectionError, connection.check_for_errors)

    def test_channel0_override_client_credentials(self):
        client_properties = {'platform': 'Atari', 'license': 'MIT'}
        channel = Channel0(FakeConnection(), client_properties)
        result = channel._client_properties()

        information = 'See https://github.com/eandersson/amqpstorm'

        self.assertIsInstance(result, dict)

        self.assertTrue(result['capabilities']['authentication_failure_close'])
        self.assertTrue(result['capabilities']['consumer_cancel_notify'])
        self.assertTrue(result['capabilities']['publisher_confirms'])
        self.assertTrue(result['capabilities']['connection.blocked'])
        self.assertTrue(result['capabilities']['basic.nack'])
        self.assertEqual(result['information'], information)
        self.assertEqual(result['platform'], 'Atari')
        self.assertEqual(result['license'], 'MIT')



import socket
import threading

import mock
from pamqp import frame as pamqp_frame
from pamqp import specification
from pamqp.specification import Basic as spec_basic

from amqpstorm import Channel
from amqpstorm import Connection
from amqpstorm import compatibility
from amqpstorm.base import MAX_CHANNELS
from amqpstorm.exception import AMQPConnectionError
from amqpstorm.io import IO
from amqpstorm.tests.utility import TestFramework


class ConnectionTests(TestFramework):
    def test_connection_with_statement(self):
        with Connection('127.0.0.1', 'guest', 'guest', lazy=True) as con:
            self.assertIsInstance(con, Connection)

    def test_connection_with_statement_when_failing(self):
        try:
            with Connection('127.0.0.1', 'guest', 'guest', lazy=True) as con:
                con.exceptions.append(AMQPConnectionError('travis-ci'))
                con.check_for_errors()
        except AMQPConnectionError as why:
            self.assertIsInstance(why, AMQPConnectionError)

        self.assertEqual(self.get_last_log(),
                         'Closing connection due to an unhandled exception: '
                         'travis-ci')

    def test_connection_server_is_blocked_default_value(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        self.assertEqual(connection.is_blocked, False)

    def test_connection_server_properties_default_value(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        self.assertEqual(connection.server_properties, {})

    def test_connection_socket_property(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection._io.socket = 'FakeSocket'
        self.assertEqual(connection.socket, 'FakeSocket')

    def test_connection_socket_none_when_closed(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        self.assertFalse(connection.socket)

    def test_connection_fileno_property(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPENING)
        io = IO(connection.parameters, [])
        io.socket = mock.Mock(name='socket', spec=socket.socket)
        connection._io = io
        io.socket.fileno.return_value = 5

        self.assertEqual(connection.fileno, 5)

    def test_connection_fileno_none_when_closed(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        self.assertIsNone(connection.fileno)

    def test_connection_close_state(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(Connection.OPEN)
        connection.close()

        self.assertTrue(connection.is_closed)

    def test_connection_open_channel_on_closed_connection(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        self.assertRaises(AMQPConnectionError, connection.channel)

    def test_connection_basic_read_buffer(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        cancel_ok_frame = spec_basic.CancelOk().marshal()

        self.assertEqual(connection._read_buffer(cancel_ok_frame), b'\x00')

    def test_connection_basic_read_buffer_in_one_go(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        data_in = (
            b'\x02\x00\x01\x00\x00\x00\x83\x00<\x00\x00\x00\x00\x00\x00\x00'
            b'\x00\x00\x0c\xe4\xc0\ntext/plain\x05utf-8\x00\x00\x00\x0e\x03'
            b'keyS\x00\x00\x00\x05value$c2ab75b3-5641-43ae-85a0-79e5ff6b1044'
            b'$9d5a550d-c91d-493d-8a71-b606f0a64080\x00\x00\x00\x00a6\xac\x87'
            b'\xce\x03\x00\x01\x00\x00\x00\x0cHello World!\xce'
        )

        data_in = connection._read_buffer(data_in)
        self.assertFalse(data_in)

    def test_connection_send_handshake(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        def on_write_to_socket(message):
            self.assertEqual(message, b'AMQP\x00\x00\t\x01')

        connection._io.write_to_socket = on_write_to_socket

        self.assertIsNone(connection._send_handshake())

    def test_connection_handle_read_buffer_none_returns_none(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        self.assertIsNone(connection._read_buffer(None))

    def test_connection_basic_handle_amqp_frame(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        payload = (
            b'\x01\x00\x00\x00\x00\x00\x0c\x00\n\x00\x1e\x00\x00\x00'
            b'\x02\x00\x00\x00<\xce'
        )

        data_in, channel_id, frame_in = connection._handle_amqp_frame(payload)

        self.assertEqual(data_in, b'')
        self.assertEqual(channel_id, 0)
        self.assertIsInstance(frame_in, specification.Connection.Tune)

    def test_connection_handle_amqp_frame_none_returns_none(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        result = connection._handle_amqp_frame('')

        self.assertEqual(result[0], '')
        self.assertIsNone(result[1])
        self.assertIsNone(result[2])

    def test_connection_handle_amqp_frame_error(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        def throw_error(*_):
            raise specification.AMQPFrameError()

        restore_func = pamqp_frame.unmarshal
        try:
            pamqp_frame.unmarshal = throw_error

            result = connection._handle_amqp_frame('travis-ci')

            self.assertEqual(result[0], 'travis-ci')
            self.assertIsNone(result[1])
            self.assertIsNone(result[2])
        finally:
            pamqp_frame.unmarshal = restore_func

        self.assertEqual(self.get_last_log(),
                         'AMQPFrameError: AMQPFrameError()')

    def test_connection_handle_unicode_error(self):
        """This test covers an unlikely issue triggered by network corruption.

            pamqp.decode._maybe_utf8 raises:
                UnicodeDecodeError: 'utf8' codec can't
                decode byte 0xc5 in position 1: invalid continuation byte

            The goal here is not to fix issues caused by network corruption,
            but rather to make sure that the exceptions raised when
            connections do fail are always predictable.

            Fail fast and reliably!

        :return:
        """
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        def throw_error(_):
            raise UnicodeDecodeError(str(), bytes(), 1, 1, str())

        restore_func = pamqp_frame.unmarshal
        try:
            pamqp_frame.unmarshal = throw_error

            result = connection._handle_amqp_frame('travis-ci')

            self.assertEqual(result[0], 'travis-ci')
            self.assertIsNone(result[1])
            self.assertIsNone(result[2])
        finally:
            pamqp_frame.unmarshal = restore_func

        self.assertEqual(self.get_last_log(),
                         "'' codec can't decode bytes in position 1-0: ")

    def test_connection_handle_value_error(self):
        """This test covers an unlikely issue triggered by network corruption.

            pamqp.decode._embedded_value raises:
                ValueError: Unknown type: b'\x13'

            The goal here is not to fix issues caused by network corruption,
            but rather to make sure that the exceptions raised when
            connections do fail are always predictable.

            Fail fast and reliably!

        :return:
        """
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        def throw_error(_):
            raise ValueError("Unknown type: b'\x13'")

        restore_func = pamqp_frame.unmarshal
        try:
            pamqp_frame.unmarshal = throw_error

            result = connection._handle_amqp_frame('travis-ci')

            self.assertEqual(result[0], 'travis-ci')
            self.assertIsNone(result[1])
            self.assertIsNone(result[2])
        finally:
            pamqp_frame.unmarshal = restore_func

        self.assertEqual(self.get_last_log(),
                         "Unknown type: b'\x13'")

    def test_connection_wait_for_connection(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPENING)
        io = IO(connection.parameters, [])
        io.socket = mock.Mock(name='socket', spec=socket.socket)
        connection._io = io

        self.assertFalse(connection.is_open)

        def set_state_to_open(conn):
            conn.set_state(conn.OPEN)

        threading.Timer(function=set_state_to_open,
                        interval=0.1, args=(connection,)).start()
        connection._wait_for_connection_state(connection.OPEN)

        self.assertTrue(connection.is_open)

    def test_connection_wait_for_connection_does_raise_on_error(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPENING)

        connection.exceptions.append(AMQPConnectionError('travis-ci'))

        self.assertRaises(
            AMQPConnectionError, connection._wait_for_connection_state,
            connection.OPEN, 0.1
        )

    def test_connection_wait_for_connection_raises_on_timeout(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', timeout=0.1,
                                lazy=True)
        connection.set_state(connection.OPENING)
        io = IO(connection.parameters, [])
        io.socket = mock.Mock(name='socket', spec=socket.socket)
        connection._io = io

        self.assertRaises(
            AMQPConnectionError,
            connection._wait_for_connection_state,
            connection.OPEN, 0.1
        )

    def test_connection_open(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        io = IO(connection.parameters, [])
        io.socket = mock.Mock(name='socket', spec=socket.socket)
        connection._io = io

        def open():
            pass

        def on_write_to_socket(_):
            connection.set_state(connection.OPEN)

        connection._io.open = open
        connection._io.write_to_socket = on_write_to_socket

        self.assertTrue(connection.is_closed)

        connection.open()

        self.assertTrue(connection.is_open)

    def test_connection_close(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)
        io = IO(connection.parameters, [])
        io.socket = mock.Mock(name='socket', spec=socket.socket)
        connection._io = io

        # Create some fake channels.
        for index in range(10):
            connection._channels[index + 1] = Channel(
                index + 1, connection, 360)

        def on_write(frame_out):
            self.assertIsInstance(frame_out, specification.Connection.Close)
            connection._channel0._close_connection_ok()

        connection._channel0._write_frame = on_write

        self.assertFalse(connection.is_closed)

        connection.close()

        # Make sure all the fake channels were closed as well.
        for index in range(10):
            self.assertNotIn(index + 1, connection._channels)

        self.assertTrue(connection.is_closed)

    def test_connection_close_when_already_closed(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)
        io = IO(connection.parameters, [])
        io.socket = mock.Mock(name='socket', spec=socket.socket)
        connection._io = io

        connection.set_state(connection.CLOSED)

        # Create some fake channels.
        for index in range(10):
            connection._channels[index + 1] = Channel(
                index + 1, connection, 360)

        def state_set(state):
            self.assertEqual(state, connection.CLOSED)

        connection.set_state = state_set

        self.assertTrue(connection.is_closed)

        connection.close()

        # Make sure all the fake channels were closed as well.
        for index in range(10):
            self.assertNotIn(index + 1, connection._channels)

        self.assertFalse(connection._channels)
        self.assertTrue(connection.is_closed)

    def test_connection_close_handles_raise_on_write(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)
        io = IO(connection.parameters, [])
        io.socket = mock.Mock(name='socket', spec=socket.socket)
        connection._io = io

        # Create some fake channels.
        for index in range(10):
            connection._channels[index + 1] = Channel(
                index + 1, connection, 360)

        def raise_on_write(_):
            raise AMQPConnectionError('travis-ci')

        connection._channel0._write_frame = raise_on_write

        self.assertFalse(connection.is_closed)

        connection.close()

        # Make sure all the fake channels were closed as well.
        for index in range(10):
            self.assertNotIn(index + 1, connection._channels)

        self.assertFalse(connection._channels)
        self.assertTrue(connection.is_closed)

    def test_connection_close_channels(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        channel_1 = Channel(1, connection, 360)
        channel_2 = Channel(2, connection, 360)
        channel_3 = Channel(3, connection, 360)

        channel_1.set_state(Channel.CLOSED)
        channel_2.set_state(Channel.CLOSED)
        channel_2.set_state(Channel.OPEN)

        connection._channels[1] = channel_1
        connection._channels[2] = channel_2
        connection._channels[3] = channel_3

        connection._close_remaining_channels()

        self.assertTrue(channel_1.is_closed)
        self.assertTrue(channel_2.is_closed)
        self.assertTrue(channel_3.is_closed)

        self.assertFalse(connection._channels)

    def test_connection_closed_on_exception(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)
        connection.exceptions.append(AMQPConnectionError('travis-ci'))

        self.assertTrue(connection.is_open)
        self.assertRaises(AMQPConnectionError, connection.check_for_errors)
        self.assertTrue(connection.is_closed)

    def test_connection_heartbeat_stopped_on_close(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)
        connection.heartbeat.start(connection.exceptions)
        connection.exceptions.append(AMQPConnectionError('travis-ci'))

        self.assertTrue(connection.heartbeat._running.is_set())

        self.assertRaises(AMQPConnectionError, connection.check_for_errors)

        self.assertFalse(connection.heartbeat._running.is_set())

    def test_connection_open_new_channel(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)

        def on_open_ok(_, frame_out):
            self.assertIsInstance(frame_out, specification.Channel.Open)
            connection._channels[1].on_frame(specification.Channel.OpenOk())

        connection.write_frame = on_open_ok

        connection.channel()

    def test_connection_get_first_channel_id(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        self.assertEqual(connection._last_channel_id, None)
        self.assertEqual(
            connection._get_next_available_channel_id(), 1
        )
        self.assertEqual(connection._last_channel_id, 1)

    def test_connection_get_channel_ids(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        self.assertEqual(connection._last_channel_id, None)
        channel = Channel(1, connection, 1)
        channel.set_state(channel.OPEN)
        connection._channels[1] = channel

        for index in range(2, 100):
            channel_id = connection._get_next_available_channel_id()
            channel = Channel(channel_id, connection, 1)
            channel.set_state(channel.OPEN)
            connection._channels[channel_id] = channel
            self.assertEqual(
                channel_id, index
            )
            self.assertEqual(connection._last_channel_id, index)

    def test_connection_avoid_conflicts_with_channel_ids(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)

        for index in compatibility.RANGE(1, 301):
            channel = Channel(index, connection, 1)
            channel.set_state(channel.OPEN)
            connection._channels[index] = channel
        for index in compatibility.RANGE(302, 65535):
            channel = Channel(index, connection, 1)
            channel.set_state(channel.OPEN)
            connection._channels[index] = channel

        channel = connection.channel(lazy=True)
        channel.set_state(channel.OPEN)
        last_channel_id = int(channel)

        self.assertEqual(
            last_channel_id, 301
        )
        self.assertEqual(connection._last_channel_id, 301)

        channel = connection.channel(lazy=True)
        channel.set_state(channel.OPEN)
        last_channel_id = int(channel)

        self.assertEqual(
            last_channel_id, 65535
        )
        self.assertEqual(connection._last_channel_id, 65535)

    def test_connection_reuse_predictable(self):
        max_channels = 1024
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection._channel0.max_allowed_channels = max_channels
        connection.set_state(connection.OPEN)

        for _ in compatibility.RANGE(max_channels):
            channel = connection.channel(lazy=True)
            channel.set_state(channel.OPEN)

        ids_to_close = [2, 8, 16, 32, 64, 128, 256, 512, 768, 1024]

        for channel_id in ids_to_close:
            connection._channels[channel_id].set_state(Channel.CLOSED)

        for _ in ids_to_close:
            channel = connection.channel(lazy=True)
            channel.set_state(channel.OPEN)
            self.assertIn(int(channel), ids_to_close)

    def test_connection_reuse_old_closed_channels(self):
        max_channels = 1024
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection._channel0.max_allowed_channels = max_channels
        connection.set_state(connection.OPEN)

        for _ in compatibility.RANGE(max_channels):
            channel = connection.channel(lazy=True)
            channel.set_state(channel.OPEN)

        ids_to_close = [2, 8, 16, 32, 64, 128, 256, 512, 768, 1024]

        for _ in range(100):
            for channel_id in ids_to_close:
                connection._channels[channel_id].set_state(Channel.CLOSED)

            for _ in ids_to_close:
                self.assertIn(
                    int(connection.channel(lazy=True)), ids_to_close
                )

    def test_connection_close_old_ids_with_channel_deleted(self):
        max_channels = 1024
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection._channel0.max_allowed_channels = max_channels
        connection.set_state(connection.OPEN)

        for _ in compatibility.RANGE(max_channels):
            channel = connection.channel(lazy=True)
            channel.set_state(channel.OPEN)

        ids_to_close = [2, 8, 16, 32, 64, 128, 256, 512, 768, 1024]

        for _ in range(100):
            for channel_id in ids_to_close:
                del connection._channels[channel_id]

            self.assertEqual(len(connection._channels),
                             max_channels - len(ids_to_close))

            for _ in ids_to_close:
                channel = connection.channel(lazy=True)
                channel.set_state(channel.OPEN)

                self.assertIn(int(channel), ids_to_close)

            self.assertEqual(len(connection._channels), max_channels)

    def test_connection_dont_reuse_when_not_closed(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)

        # Open Channel 1
        channel1 = connection.channel(lazy=True)
        channel1.set_state(Channel.OPEN)

        # Open Channel 2 (but leave it closed)
        channel2 = connection.channel(lazy=True)
        channel2.set_state(Channel.CLOSED)

        self.assertEqual(2, int(channel2))

        # Channel 2 is re-used
        channel2 = connection.channel(lazy=True)
        channel2.set_state(Channel.OPEN)

        self.assertEqual(2, int(channel2))

        # New channel id is used for the third channel
        channel3 = connection.channel(lazy=True)
        channel3.set_state(Channel.OPEN)

        self.assertEqual(3, int(channel3))

        # We should now have three channels.
        self.assertEqual(3, len(connection._channels))

    def test_connection_open_many_channels(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)

        for index in compatibility.RANGE(MAX_CHANNELS - 1):
            channel = connection.channel(lazy=True)
            channel.set_state(channel.OPEN)
            self.assertEqual(int(channel), index + 1)

    def test_connection_maximum_channels_reached(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection.set_state(connection.OPEN)

        for index in compatibility.RANGE(MAX_CHANNELS):
            channel_id = connection._get_next_available_channel_id()
            channel = Channel(channel_id, connection, 1)
            channel.set_state(channel.OPEN)
            connection._channels[channel_id] = channel

            self.assertEqual(connection._last_channel_id, index + 1)

        self.assertRaisesRegex(
            AMQPConnectionError,
            'reached the maximum number of channels %d' % MAX_CHANNELS,
            connection.channel, lazy=True)

    def test_connection_cleanup_one_channel(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection._channels[1] = Channel(1, connection, 0.1)

        connection._cleanup_channel(1)

        self.assertFalse(connection._channels)

    def test_connection_cleanup_multiple_channels(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)

        for index in range(1, 10):
            connection._channels[index] = Channel(index, connection, 0.1)

        for index in range(1, 10):
            connection._cleanup_channel(index)

        self.assertFalse(connection._channels)

    def test_connection_cleanup_channel_does_not_exist(self):
        connection = Connection('127.0.0.1', 'guest', 'guest', lazy=True)
        connection._channels[1] = Channel(1, connection, 0.1)

        connection._cleanup_channel(2)

        self.assertEqual(len(connection._channels), 1)



import requests

from amqpstorm.management.http_client import ApiError
from amqpstorm.management.http_client import HTTPClient
from amqpstorm.tests.utility import TestFramework


class FakeResponse(object):
    """Fake Requests Response for Unit-Testing."""

    def __init__(self, status_code=200, json=None, raises=None):
        self.status_code = status_code
        self._json = json
        self._raises = raises

    def raise_for_status(self):
        if not self._raises:
            return
        raise self._raises

    def json(self):
        if self._raises:
            raise self._raises
        return self._json


class ApiHTTPTests(TestFramework):
    def test_api_valid_json(self):
        fake_payload = {
            'hello': 'travis-ci'
        }

        response = HTTPClient._get_json_output(FakeResponse(json=fake_payload))

        self.assertEqual(response, {
            'hello': 'travis-ci'
        })

    def test_api_invalid_json(self):
        response = HTTPClient._get_json_output(
            FakeResponse(raises=ValueError)
        )
        self.assertIsNone(response)

    def test_api_http_standard_http_error(self):
        fake_payload = {
            'error': 'travis-ci'
        }

        self.assertRaisesRegex(
            ApiError, 'travis-ci',
            HTTPClient._check_for_errors,
            FakeResponse(raises=requests.HTTPError('travis-ci')),
            fake_payload
        )

    def test_api_http_non_standard_http_error(self):
        fake_payload = {
            'error': 'travis-ci'
        }

        self.assertRaisesRegex(
            ApiError, 'travis-ci',
            HTTPClient._check_for_errors,
            FakeResponse(),
            fake_payload
        )



"""AMQPStorm Exception."""

AMQP_ERROR_MAPPING = {
    311: ('CONTENT-TOO-LARGE',
          'The client attempted to transfer content larger than the '
          'server could accept at the present time. The client may '
          'retry  at a later time.'),
    312: ('NO-ROUTE', 'Undocumented AMQP Soft Error'),
    313: ('NO-CONSUMERS',
          'When the exchange cannot deliver to a consumer when the '
          'immediate flag is set. As a result of pending data on '
          'the queue or the absence of any consumers of the queue.'),
    320: ('CONNECTION-FORCED',
          'An operator intervened to close the connection for some reason. '
          'The client may retry at some later date.'),
    402: ('INVALID-PATH',
          'The client tried to work with an unknown virtual host.'),
    403: ('ACCESS-REFUSED',
          'The client attempted to work with a server entity to which '
          'has no access due to security settings.'),
    404: ('NOT-FOUND',
          'The client attempted to work with a server '
          'entity that does not exist.'),
    405: ('RESOURCE-LOCKED',
          'The client attempted to work with a server entity to which it '
          'has no access because another client is working with it.'),
    406: ('PRECONDITION-FAILED',
          'The client requested a method that was not '
          'allowed because some precondition failed.'),
    501: ('FRAME-ERROR',
          'The sender sent a malformed frame that the recipient could '
          'not decode. This strongly implies a programming error in '
          'the sending peer.'),
    502: ('SYNTAX-ERROR',
          'The sender sent a frame that contained illegal values for '
          'one or more fields. This strongly implies a programming '
          'error in the sending peer.'),
    503: ('COMMAND-INVALID',
          'The client sent an invalid sequence of frames, attempting to '
          'perform an operation that was considered invalid by the server. '
          'This usually implies a programming error in the client.'),
    504: ('CHANNEL-ERROR',
          'The client attempted to work with a channel that had not '
          'been correctly opened. This most likely indicates a '
          'fault in the client layer.'),
    505: ('UNEXPECTED-FRAME',
          'The peer sent a frame that was not expected, usually in the '
          'context of a content header and body. This strongly '
          'indicates a fault in the peer\'s content processing.'),
    506: ('RESOURCE-ERROR',
          'The server could not complete the method because it lacked '
          'sufficient resources. This may be due to the client '
          'creating too many of some type of entity.'),
    530: ('NOT-ALLOWED',
          'The client tried to work with some entity in a manner '
          'that is prohibited by the server, due to security '
          'settings or by some other criteria.'),
    540: ('NOT-IMPLEMENTED',
          'The client tried to use functionality that is '
          'notimplemented in the server.'),
    541: ('INTERNAL-ERROR',
          'The server could not complete the method because of an '
          'internal error. The server may require intervention by '
          'an operator in order to resume normal operations.')
}


class AMQPError(IOError):
    """General AMQP Error.

    Exceptions raised by AMQPStorm are mapped based to the
    AMQP 0.9.1 specifications (when applicable).

    e.g.
    ::

        except AMQPChannelError as why:
            if why.error_code == 312:
                self.channel.queue.declare(queue_name)
    """
    _documentation = None
    _error_code = None
    _error_type = None

    @property
    def documentation(self):
        """AMQP Documentation string."""
        return self._documentation or bytes()

    @property
    def error_code(self):
        """AMQP Error Code - A 3-digit reply code."""
        return self._error_code

    @property
    def error_type(self):
        """AMQP Error Type e.g. NOT-FOUND."""
        return self._error_type

    def __init__(self, *args, **kwargs):
        self._error_code = kwargs.pop('reply_code', None)
        super(AMQPError, self).__init__(*args, **kwargs)
        if self._error_code not in AMQP_ERROR_MAPPING:
            return
        self._error_type = AMQP_ERROR_MAPPING[self._error_code][0]
        self._documentation = AMQP_ERROR_MAPPING[self._error_code][1]


class AMQPConnectionError(AMQPError):
    """AMQP Connection Error."""
    pass


class AMQPChannelError(AMQPError):
    """AMQP Channel Error."""
    pass


class AMQPMessageError(AMQPChannelError):
    """AMQP Message Error."""
    pass


class AMQPInvalidArgument(AMQPError):
    """AMQP Argument Error."""



import requests.api
from requests.auth import HTTPBasicAuth

from amqpstorm.compatibility import urlparse
from amqpstorm.management.exception import ApiConnectionError
from amqpstorm.management.exception import ApiError


class HTTPClient(object):
    def __init__(self, api_url, username, password, verify, cert, timeout):
        self.session = requests.Session()
        self.session.verify = verify
        self.session.cert = cert
        self._auth = HTTPBasicAuth(username, password)
        self._base_url = api_url
        self._timeout = timeout

    def get(self, path, payload=None, headers=None):
        """HTTP GET operation.

        :param path: URI Path
        :param payload: HTTP Body
        :param headers: HTTP Headers

        :raises ApiError: Raises if the remote server encountered an error.
        :raises ApiConnectionError: Raises if there was a connectivity issue.

        :return: Response
        """
        return self._request('get', path, payload, headers)

    def list(self, path, name=None, page_size=None, use_regex=False):
        """List operation (e.g. queue list).

        :param path: URI Path
        :param name: Filter by name, for example queue name, exchange name etc
        :param use_regex: Enables regular expression for the param name
        :param page_size: Number of elements per page

        :raises ApiError: Raises if the remote server encountered an error.
        :raises ApiConnectionError: Raises if there was a connectivity issue.

        :return: Response
        """
        params = dict()
        if name is not None:
            params['name'] = name
        if use_regex:
            if isinstance(use_regex, bool):
                use_regex = str(use_regex)
            params['use_regex'] = use_regex.lower()

        if page_size is None:
            return self._request('get', path, params=params)

        results = list()
        current_page = 1
        params['page'] = current_page
        params['page_size'] = page_size
        params['pagination'] = True
        first_result = self._request('get', path, params=params)
        num_pages = first_result['page_count']
        current_page = first_result.get('page', 1)
        results.extend(first_result['items'])

        while num_pages > current_page:
            params['page'] = current_page + 1
            next_result = self._request('get', path, params=params)
            current_page = next_result['page']
            num_pages = next_result['page_count']
            items = next_result.get('items')
            if not items:
                break
            results.extend(items)

        return results

    def post(self, path, payload=None, headers=None):
        """HTTP POST operation.

        :param path: URI Path
        :param payload: HTTP Body
        :param headers: HTTP Headers

        :raises ApiError: Raises if the remote server encountered an error.
        :raises ApiConnectionError: Raises if there was a connectivity issue.

        :return: Response
        """
        return self._request('post', path, payload, headers)

    def delete(self, path, payload=None, headers=None):
        """HTTP DELETE operation.

        :param path: URI Path
        :param payload: HTTP Body
        :param headers: HTTP Headers

        :raises ApiError: Raises if the remote server encountered an error.
        :raises ApiConnectionError: Raises if there was a connectivity issue.

        :return: Response
        """
        return self._request('delete', path, payload, headers)

    def put(self, path, payload=None, headers=None):
        """HTTP PUT operation.

        :param path: URI Path
        :param payload: HTTP Body
        :param headers: HTTP Headers

        :raises ApiError: Raises if the remote server encountered an error.
        :raises ApiConnectionError: Raises if there was a connectivity issue.

        :return: Response
        """
        return self._request('put', path, payload, headers)

    def _request(self, method, path, payload=None, headers=None, params=None):
        """HTTP operation.

        :param method: Operation type (e.g. post)
        :param path: URI Path
        :param payload: HTTP Body
        :param headers: HTTP Headers
        :param params: HTTP Parameters

        :raises ApiError: Raises if the remote server encountered an error.
        :raises ApiConnectionError: Raises if there was a connectivity issue.

        :return: Response
        """
        url = urlparse.urljoin(self._base_url, 'api/%s' % path)
        headers = headers or {}
        headers['content-type'] = 'application/json'
        try:
            response = self.session.request(
                method, url,
                auth=self._auth,
                data=payload,
                headers=headers,
                timeout=self._timeout,
                params=params,
            )
        except requests.RequestException as why:
            raise ApiConnectionError(str(why))

        json_response = self._get_json_output(response)
        self._check_for_errors(response, json_response)
        return json_response

    @staticmethod
    def _get_json_output(response):
        """Get JSON output from the HTTP response.

        :param requests.Response response:

        :return: Json payload
        """
        try:
            content = response.json()
        except ValueError:
            content = None
        return content

    @staticmethod
    def _check_for_errors(response, json_response):
        """Check payload for errors.

        :param response: HTTP response
        :param json_response: Json response

        :raises ApiError: Raises if the remote server encountered an error.

        :return:
        """
        status_code = response.status_code
        try:
            response.raise_for_status()
        except requests.HTTPError as why:
            raise ApiError(str(why), reply_code=status_code)
        if isinstance(json_response, dict) and 'error' in json_response:
            raise ApiError(json_response['error'], reply_code=status_code)



"""AMQPStorm Channel.Tx."""

import logging

from pamqp import specification

from amqpstorm.base import Handler

LOGGER = logging.getLogger(__name__)


class Tx(Handler):
    """RabbitMQ Transactions.

        Server local transactions, in which the server will buffer published
        messages until the client commits (or rollback) the messages.

    """
    __slots__ = ['_tx_active']

    def __init__(self, channel):
        self._tx_active = True
        super(Tx, self).__init__(channel)

    def __enter__(self):
        self.select()
        return self

    def __exit__(self, exception_type, exception_value, _):
        if exception_type:
            LOGGER.warning(
                'Leaving Transaction on exception: %s',
                exception_value
            )
            if self._tx_active:
                self.rollback()
            return
        if self._tx_active:
            self.commit()

    def select(self):
        """Enable standard transaction mode.

            This will enable transaction mode on the channel. Meaning that
            messages will be kept in the remote server buffer until such a
            time that either commit or rollback is called.

        :return:
        """
        self._tx_active = True
        return self._channel.rpc_request(specification.Tx.Select())

    def commit(self):
        """Commit the current transaction.

            Commit all messages published during the current transaction
            session to the remote server.

            A new transaction session starts as soon as the command has
            been executed.

        :return:
        """
        self._tx_active = False
        return self._channel.rpc_request(specification.Tx.Commit())

    def rollback(self):
        """Abandon the current transaction.

            Rollback all messages published during the current transaction
            session to the remote server.

            Note that all messages published during this transaction session
            will be lost, and will have to be published again.

            A new transaction session starts as soon as the command has
            been executed.

        :return:
        """
        self._tx_active = False
        return self._channel.rpc_request(specification.Tx.Rollback())



import socket
import ssl

import mock

import amqpstorm.io
from amqpstorm.exception import AMQPConnectionError
from amqpstorm.io import IO
from amqpstorm.tests.utility import FakeConnection
from amqpstorm.tests.utility import TestFramework


class IOTests(TestFramework):

    def test_io_socket_close(self):
        connection = FakeConnection()
        io = IO(connection.parameters)
        io.socket = mock.Mock(name='socket', spec=socket.socket)
        io.close()

        self.assertIsNone(io.socket)

    def test_io_use_ssl_false(self):
        connection = FakeConnection()
        io = IO(connection.parameters)

        self.assertFalse(io.use_ssl)

    def test_io_use_ssl_true(self):
        connection = FakeConnection()
        connection.parameters['ssl'] = True
        io = IO(connection.parameters)

        self.assertTrue(io.use_ssl)

    def test_io_create_socket(self):
        connection = FakeConnection()
        io = IO(connection.parameters)

        self.assertFalse(io.use_ssl)

        addresses = io._get_socket_addresses()
        sock_address_tuple = addresses[0]
        sock = io._create_socket(socket_family=sock_address_tuple[0])

        if hasattr(socket, 'socket'):
            self.assertIsInstance(sock, socket.socket)

    def test_io_create_ssl_socket(self):
        connection = FakeConnection()
        connection.parameters['ssl'] = True
        io = IO(connection.parameters)

        self.assertTrue(io.use_ssl)

        addresses = io._get_socket_addresses()
        sock_address_tuple = addresses[0]
        sock = io._create_socket(socket_family=sock_address_tuple[0])

        if hasattr(socket, 'socket'):
            self.assertIsInstance(sock, socket.socket)
        if hasattr(ssl, 'SSLSocket'):
            self.assertIsInstance(sock, ssl.SSLSocket)

    def test_io_get_socket_address(self):
        connection = FakeConnection()
        connection.parameters['hostname'] = '127.0.0.1'
        connection.parameters['port'] = 5672
        io = IO(connection.parameters)
        addresses = io._get_socket_addresses()
        sock_address_tuple = addresses[0]

        self.assertEqual(sock_address_tuple[4],
                         ('127.0.0.1', 5672))

    def test_io_simple_receive(self):
        connection = FakeConnection()
        io = IO(connection.parameters)

        self.assertFalse(io.use_ssl)

        io.socket = mock.Mock(name='socket', spec=socket.socket)
        io.socket.recv.return_value = '12345'

        self.assertEqual(io._receive(), '12345')

    def test_io_simple_ssl_receive(self):
        connection = FakeConnection()
        connection.parameters['ssl'] = True
        io = IO(connection.parameters)

        self.assertTrue(io.use_ssl)

        if hasattr(ssl, 'SSLObject'):
            io.socket = mock.Mock(name='socket', spec=ssl.SSLObject)
        elif hasattr(ssl, 'SSLSocket'):
            io.socket = mock.Mock(name='socket', spec=ssl.SSLSocket)

        io.socket.read.return_value = '12345'

        self.assertEqual(io._receive(), '12345')

    def test_io_simple_send_zero_bytes_sent(self):
        connection = FakeConnection()

        io = IO(connection.parameters, exceptions=connection.exceptions)
        io.socket = mock.Mock(name='socket', spec=socket.socket)
        io.socket.send.return_value = 0
        io.write_to_socket(self.message)

        self.assertRaisesRegex(
            AMQPConnectionError,
            'connection/socket error',
            connection.check_for_errors
        )

    def test_io_set_ssl_context(self):
        connection = FakeConnection()
        connection.parameters['ssl_options'] = {
            'context': ssl.create_default_context(),
            'server_hostname': 'localhost',
        }

        io = IO(connection.parameters)
        self.assertTrue(io._ssl_wrap_socket(socket.socket()))

    def test_io_set_ssl_verify_req(self):
        connection = FakeConnection()
        connection.parameters['ssl_options'] = {
            'verify_mode': 'required'
        }

        io = IO(connection.parameters)
        sock = io._ssl_wrap_socket(socket.socket())
        self.assertEqual(sock.context.verify_mode, ssl.CERT_REQUIRED)

    def test_io_set_ssl_context_no_hostname_provided(self):
        connection = FakeConnection()
        connection.parameters['ssl_options'] = {
            'context': ssl.create_default_context(),
        }

        io = IO(connection.parameters)
        self.assertRaises(ValueError, io._ssl_wrap_socket, socket.socket())

    def test_io_has_ipv6(self):
        restore_func = socket.getaddrinfo

        def mock_getaddrinfo(hostname, port, family, socktype):
            return [hostname, port, family, socktype]

        try:
            amqpstorm.io.socket.getaddrinfo = mock_getaddrinfo
            connection = FakeConnection()
            connection.parameters['hostname'] = 'localhost'
            connection.parameters['port'] = 1234
            parameters = connection.parameters
            io = IO(parameters)

            result = io._get_socket_addresses()
            self.assertEqual(result[2], socket.AF_UNSPEC)
            self.assertEqual(result[3], socket.SOCK_STREAM)
        finally:
            amqpstorm.io.socket.getaddrinfo = restore_func

    def test_io_has_ipv6_is_false(self):
        restore_func = socket.getaddrinfo
        restore_has_ipv6 = amqpstorm.io.socket.has_ipv6

        def mock_getaddrinfo(hostname, port, family, socktype):
            return [hostname, port, family, socktype]

        try:
            amqpstorm.io.socket.getaddrinfo = mock_getaddrinfo
            amqpstorm.io.socket.has_ipv6 = False
            connection = FakeConnection()
            connection.parameters['hostname'] = 'localhost'
            connection.parameters['port'] = 1234
            parameters = connection.parameters
            io = IO(parameters)

            result = io._get_socket_addresses()
            self.assertEqual(result[2], socket.AF_INET)
            self.assertEqual(result[3], socket.SOCK_STREAM)
        finally:
            amqpstorm.io.socket.getaddrinfo = restore_func
            amqpstorm.io.socket.has_ipv6 = restore_has_ipv6



"""AMQPStorm Base."""

import locale

from amqpstorm.compatibility import is_string
from amqpstorm.exception import AMQPChannelError

AUTH_MECHANISM = 'PLAIN'
IDLE_WAIT = 0.01
LOCALE = locale.getlocale()[0] or 'en_US'
MAX_FRAME_SIZE = 131072
MAX_CHANNELS = 65535


class Stateful(object):
    """Stateful implementation."""
    CLOSED = 0
    CLOSING = 1
    OPENING = 2
    OPEN = 3

    def __init__(self):
        self._state = self.CLOSED
        self._exceptions = []

    def set_state(self, state):
        """Set State.

        :param int state:
        :return:
        """
        self._state = state

    @property
    def current_state(self):
        """Get the State.

        :rtype: int
        """
        return self._state

    @property
    def is_closed(self):
        """Is Closed?

        :rtype: bool
        """
        return self._state == self.CLOSED

    @property
    def is_closing(self):
        """Is Closing?

        :rtype: bool
        """
        return self._state == self.CLOSING

    @property
    def is_opening(self):
        """Is Opening?

        :rtype: bool
        """
        return self._state == self.OPENING

    @property
    def is_open(self):
        """Is Open?

        :rtype: bool
        """
        return self._state == self.OPEN

    @property
    def exceptions(self):
        """Stores all exceptions thrown by this instance.

            This is useful for troubleshooting, and is used internally
            to check the health of the connection.

        :rtype: list
        """
        return self._exceptions


class BaseChannel(Stateful):
    """Channel base class."""
    __slots__ = [
        '_channel_id', '_consumer_tags'
    ]

    def __init__(self, channel_id):
        super(BaseChannel, self).__init__()
        self._consumer_tags = []
        self._channel_id = channel_id

    @property
    def channel_id(self):
        """Get Channel id.

        :rtype: int
        """
        return self._channel_id

    @property
    def consumer_tags(self):
        """Get a list of consumer tags.

        :rtype: list
        """
        return self._consumer_tags

    def add_consumer_tag(self, tag):
        """Add a Consumer tag.

        :param str tag: Consumer tag.
        :return:
        """
        if not is_string(tag):
            raise AMQPChannelError('consumer tag needs to be a string')
        if tag not in self._consumer_tags:
            self._consumer_tags.append(tag)

    def remove_consumer_tag(self, tag=None):
        """Remove a Consumer tag.

            If no tag is specified, all all tags will be removed.

        :param str,None tag: Consumer tag.
        :return:
        """
        if tag is not None:
            if tag in self._consumer_tags:
                self._consumer_tags.remove(tag)
        else:
            self._consumer_tags = []


class BaseMessage(object):
    """Message base class.

    :param Channel channel: AMQPStorm Channel
    :param str,unicode body: Message body
    :param dict method: Message method
    :param dict properties: Message properties
    :param bool auto_decode: This is not implemented in the base message class.
    """
    __slots__ = [
        '_auto_decode', '_body', '_channel', '_method', '_properties'
    ]

    def __init__(self, channel, body=None, method=None, properties=None,
                 auto_decode=None):
        self._auto_decode = auto_decode
        self._channel = channel
        self._body = body
        self._method = method
        self._properties = properties or {}

    def __iter__(self):
        for attribute in ['_body', '_channel', '_method', '_properties']:
            yield attribute[1::], getattr(self, attribute)

    def to_dict(self):
        """Message to Dictionary.

        :rtype: dict
        """
        return {
            'body': self._body,
            'method': self._method,
            'properties': self._properties,
            'channel': self._channel
        }

    def to_tuple(self):
        """Message to Tuple.

        :rtype: tuple
        """
        return self._body, self._channel, self._method, self._properties


class Handler(object):
    """Operations Handler (e.g. Queue, Exchange)"""
    __slots__ = [
        '_channel'
    ]

    def __init__(self, channel):
        self._channel = channel



import time

from amqpstorm import Channel
from amqpstorm.tests.functional.utility import TestFunctionalFramework
from amqpstorm.tests.functional.utility import setup


class LegacyFunctionalTests(TestFunctionalFramework):
    def configure(self):
        self.disable_logging_validation()

    @setup(queue=True)
    def test_functional_start_stop_consumer_tuple(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()

        for _ in range(5):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)

        # Store and inbound messages.
        inbound_messages = []

        def on_message(body, channel, method, properties):
            self.assertIsInstance(body, (bytes, str))
            self.assertIsInstance(channel, Channel)
            self.assertIsInstance(properties, dict)
            self.assertIsInstance(method, dict)
            inbound_messages.append(body)
            if len(inbound_messages) >= 5:
                channel.stop_consuming()

        self.channel.basic.consume(callback=on_message,
                                   queue=self.queue_name,
                                   no_ack=True)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        self.channel.start_consuming(to_tuple=True)

        # Make sure all five messages were downloaded.
        self.assertEqual(len(inbound_messages), 5)

    @setup(queue=True)
    def test_functional_publish_and_consume_five_messages_tuple(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()

        for _ in range(5):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)

        # Store and inbound messages.
        inbound_messages = []

        def on_message(body, channel, method, properties):
            self.assertEqual(body, self.message.encode('utf-8'))
            self.assertIsInstance(body, (bytes, str))
            self.assertIsInstance(channel, Channel)
            self.assertIsInstance(properties, dict)
            self.assertIsInstance(method, dict)
            inbound_messages.append(body)

        self.channel.basic.consume(callback=on_message,
                                   queue=self.queue_name,
                                   no_ack=True)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        self.channel.process_data_events(to_tuple=True)

        # Make sure all five messages were downloaded.
        self.assertEqual(len(inbound_messages), 5)

    @setup(queue=True)
    def test_functional_generator_consume(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        for _ in range(5):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)
        self.channel.basic.consume(queue=self.queue_name,
                                   no_ack=True)
        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        # Store and inbound messages.
        inbound_messages = []
        for message in self.channel.build_inbound_messages(
                break_on_empty=True,
                to_tuple=True):
            self.assertIsInstance(message, tuple)
            self.assertIsInstance(message[0], bytes)
            self.assertIsInstance(message[1], Channel)
            self.assertIsInstance(message[2], dict)
            self.assertIsInstance(message[3], dict)
            inbound_messages.append(message)

        # Make sure all five messages were downloaded.
        self.assertEqual(len(inbound_messages), 5)

    @setup(queue=True)
    def test_functional_publish_and_get_five_messages(self):
        self.channel.queue.declare(self.queue_name)

        # Publish 5 Messages.
        for _ in range(5):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        # Get 5 messages.
        for _ in range(5):
            payload = self.channel.basic.get(self.queue_name, to_dict=True)
            self.assertIsInstance(payload, dict)



import codecs
import os.path

from setuptools import find_packages
from setuptools import setup


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name='AMQPStorm',
    python_requires='>=2.7',
    version=get_version('amqpstorm/__init__.py'),
    description='Thread-safe Python RabbitMQ Client & Management library.',
    long_description=open('README.rst').read(),
    author='Erik Olof Gunnar Andersson',
    author_email='me@eandersson.net',
    include_package_data=True,
    packages=find_packages(),
    license='MIT License',
    url='https://www.amqpstorm.io',
    install_requires=['pamqp>=2.0.0,<3.0'],
    extras_require={
        'management': ['requests>2'],
        'pool': ['amqpstorm-pool']
    },
    package_data={'': ['README.rst', 'LICENSE', 'CHANGELOG.rst']},
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Communications',
        'Topic :: Internet',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: System :: Networking'
    ]
)



from amqpstorm.exception import AMQPError

from amqpstorm.tests.utility import TestFramework


class ExceptionTests(TestFramework):
    def test_exception_documentation_matching(self):
        exception = AMQPError('travis-ci', reply_code=312)

        self.assertEqual(str(exception), 'travis-ci')

        self.assertEqual(exception.documentation,
                         'Undocumented AMQP Soft Error')

    def test_exception_error_type_matching(self):
        exception = AMQPError('travis-ci', reply_code=404)

        self.assertEqual(str(exception), 'travis-ci')

        self.assertEqual(exception.error_type,
                         'NOT-FOUND')

    def test_exception_error_code_matching(self):
        exception = AMQPError('travis-ci', reply_code=406)

        self.assertEqual(str(exception), 'travis-ci')

        self.assertEqual(exception.error_code, 406)

    def test_exception_unknown_error_code(self):
        exception = AMQPError('travis-ci', reply_code=123)

        self.assertEqual(str(exception), 'travis-ci')
        self.assertEqual(exception.error_code, 123)

        self.assertFalse(exception.error_type)
        self.assertFalse(exception.documentation)

    def test_exception_no_error_code(self):
        exception = AMQPError('travis-ci')

        self.assertEqual(str(exception), 'travis-ci')

        self.assertFalse(exception.error_type)
        self.assertFalse(exception.error_code)
        self.assertFalse(exception.documentation)



"""Python 2/3 Compatibility layer."""

import sys

try:
    import ssl
except ImportError:
    ssl = None

try:
    import simplejson as json  # noqa
except ImportError:
    import json  # noqa

try:
    import urlparse  # noqa
except ImportError:
    import urllib.parse as urlparse  # noqa

try:
    from urllib import quote  # noqa
except ImportError:
    from urllib.parse import quote  # noqa

PYTHON3 = sys.version_info >= (3, 0, 0)

if PYTHON3:
    RANGE = range
else:
    RANGE = xrange


class DummyException(Exception):
    """
    Never raised by anything.

    This is used in except blocks if the intended
    exception cannot be imported.
    """


SSL_CERT_MAP = {}
SSL_VERSIONS = {}
SSL_OPTIONS = [
    'keyfile',
    'certfile',
    'cert_reqs',
    'ssl_version',
    'ca_certs',
    'server_hostname',
]


def get_default_ssl_version():
    """Get the highest support TLS version, if none is available, return None.

    :rtype: bool,None
    """
    if hasattr(ssl, 'PROTOCOL_TLSv1_2'):
        return ssl.PROTOCOL_TLSv1_2
    elif hasattr(ssl, 'PROTOCOL_TLSv1_1'):
        return ssl.PROTOCOL_TLSv1_1
    elif hasattr(ssl, 'PROTOCOL_TLSv1'):
        return ssl.PROTOCOL_TLSv1
    return None


DEFAULT_SSL_VERSION = get_default_ssl_version()
SSL_SUPPORTED = DEFAULT_SSL_VERSION is not None
if SSL_SUPPORTED:
    if hasattr(ssl, 'PROTOCOL_TLSv1_2'):
        SSL_VERSIONS['protocol_tlsv1_2'] = ssl.PROTOCOL_TLSv1_2
    if hasattr(ssl, 'PROTOCOL_TLSv1_1'):
        SSL_VERSIONS['protocol_tlsv1_1'] = ssl.PROTOCOL_TLSv1_1
    SSL_VERSIONS['protocol_tlsv1'] = ssl.PROTOCOL_TLSv1

    SSL_CERT_MAP = {
        'cert_none': ssl.CERT_NONE,
        'cert_optional': ssl.CERT_OPTIONAL,
        'cert_required': ssl.CERT_REQUIRED
    }
    SSLWantReadError = ssl.SSLWantReadError
else:
    SSLWantReadError = DummyException


def is_string(obj):
    """Is this a string.

    :param object obj:
    :rtype: bool
    """
    if PYTHON3:
        str_type = (bytes, str)
    else:
        str_type = (bytes, str, unicode)
    return isinstance(obj, str_type)


def is_integer(obj):
    """Is this an integer.

    :param object obj:
    :return:
    """
    if PYTHON3:
        return isinstance(obj, int)
    return isinstance(obj, (int, long))


def is_unicode(obj):
    """Is this a unicode string.

        This always returns False if running Python 3.x.

    :param object obj:
    :rtype: bool
    """
    if PYTHON3:
        return False
    return isinstance(obj, unicode)


def try_utf8_decode(value):
    """Try to decode an object.

    :param value:
    :return:
    """
    if not value or not is_string(value):
        return value
    elif PYTHON3 and not isinstance(value, bytes):
        return value
    elif not PYTHON3 and not isinstance(value, unicode):
        return value

    try:
        return value.decode('utf-8')
    except UnicodeDecodeError:
        pass

    return value


def patch_uri(uri):
    """If a custom uri schema is used with python 2.6 (e.g. amqps),
    it will ignore some of the parsing logic.

        As a work-around for this we change the amqp/amqps schema
        internally to use http/https.

    :param str uri: AMQP Connection string
    :rtype: str
    """
    index = uri.find(':')
    if uri[:index] == 'amqps':
        uri = uri.replace('amqps', 'https', 1)
    elif uri[:index] == 'amqp':
        uri = uri.replace('amqp', 'http', 1)
    return uri



"""AMQPStorm Connection."""

import logging
import threading
import time
from time import sleep

from pamqp import exceptions as pamqp_exception
from pamqp import frame as pamqp_frame
from pamqp import header as pamqp_header
from pamqp import specification

from amqpstorm import compatibility
from amqpstorm.base import IDLE_WAIT
from amqpstorm.base import Stateful
from amqpstorm.channel import Channel
from amqpstorm.channel0 import Channel0
from amqpstorm.exception import AMQPConnectionError
from amqpstorm.exception import AMQPInvalidArgument
from amqpstorm.heartbeat import Heartbeat
from amqpstorm.io import EMPTY_BUFFER
from amqpstorm.io import IO

LOGGER = logging.getLogger(__name__)

DEFAULT_HEARTBEAT_TIMEOUT = 60
DEFAULT_SOCKET_TIMEOUT = 10
DEFAULT_VIRTUAL_HOST = '/'


class Connection(Stateful):
    """RabbitMQ Connection.

    e.g.
    ::

        import amqpstorm
        connection = amqpstorm.Connection('localhost', 'guest', 'guest')

    Using a SSL Context:
    ::

        import ssl
        import amqpstorm
        ssl_options = {
            'context': ssl.create_default_context(cafile='ca_certificate.pem'),
            'server_hostname': 'rmq.amqpstorm.io',
            'check_hostname': True,        # New 2.8.0, default is False
            'verify_mode': 'required',     # New 2.8.0, default is 'none'
        }
        connection = amqpstorm.Connection(
            'rmq.amqpstorm.io', 'guest', 'guest', port=5671,
            ssl=True, ssl_options=ssl_options
        )

    :param str hostname: Hostname
    :param str username: Username
    :param str password: Password
    :param int port: Server port
    :param str virtual_host: Virtual host
    :param int heartbeat: RabbitMQ Heartbeat timeout
    :param int,float timeout: Socket timeout
    :param bool ssl: Enable SSL
    :param dict ssl_options: SSL kwargs
    :param dict client_properties: None or dict of client properties
    :param bool lazy: Lazy initialize the connection

    :raises AMQPConnectionError: Raises if the connection
                                 encountered an error.
    """
    __slots__ = [
        'heartbeat', 'parameters', '_channel0', '_channels', '_io'
    ]

    def __init__(self, hostname, username, password, port=5672, **kwargs):
        super(Connection, self).__init__()
        self.lock = threading.RLock()
        self.parameters = {
            'hostname': hostname,
            'username': username,
            'password': password,
            'port': port,
            'virtual_host': kwargs.get('virtual_host', DEFAULT_VIRTUAL_HOST),
            'heartbeat': kwargs.get('heartbeat', DEFAULT_HEARTBEAT_TIMEOUT),
            'timeout': kwargs.get('timeout', DEFAULT_SOCKET_TIMEOUT),
            'ssl': kwargs.get('ssl', False),
            'ssl_options': kwargs.get('ssl_options', {}),
            'client_properties': kwargs.get('client_properties', {})
        }
        self._validate_parameters()
        self._io = IO(self.parameters, exceptions=self._exceptions,
                      on_read_impl=self._read_buffer)
        self._channel0 = Channel0(self, self.parameters['client_properties'])
        self._channels = {}
        self._last_channel_id = None
        self.heartbeat = Heartbeat(self.parameters['heartbeat'],
                                   self._channel0.send_heartbeat)
        if not kwargs.get('lazy', False):
            self.open()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, _):
        if exception_type:
            message = 'Closing connection due to an unhandled exception: %s'
            LOGGER.warning(message, exception_value)
        self.close()

    @property
    def channels(self):
        """Returns a dictionary of the Channels currently available.

        :rtype: dict
        """
        return self._channels

    @property
    def fileno(self):
        """Returns the Socket File number.

        :rtype: integer,None
        """
        if not self._io.socket:
            return None
        return self._io.socket.fileno()

    @property
    def is_blocked(self):
        """Is the connection currently being blocked from publishing by
        the remote server.

        :rtype: bool
        """
        return self._channel0.is_blocked

    @property
    def max_allowed_channels(self):
        """Returns the maximum allowed channels for the connection.

        :rtype: int
        """
        return self._channel0.max_allowed_channels

    @property
    def max_frame_size(self):
        """Returns the maximum allowed frame size for the connection.

        :rtype: int
        """
        return self._channel0.max_frame_size

    @property
    def server_properties(self):
        """Returns the RabbitMQ Server Properties.

        :rtype: dict
        """
        return self._channel0.server_properties

    @property
    def socket(self):
        """Returns an instance of the Socket used by the Connection.

        :rtype: socket.socket
        """
        return self._io.socket

    def channel(self, rpc_timeout=60, lazy=False):
        """Open a Channel.

        :param int rpc_timeout: Timeout before we give up waiting for an RPC
                                response from the server.

        :raises AMQPInvalidArgument: Invalid Parameters
        :raises AMQPChannelError: Raises if the channel encountered an error.
        :raises AMQPConnectionError: Raises if the connection
                                     encountered an error.

        :rtype: amqpstorm.Channel
        """
        LOGGER.debug('Opening a new Channel')
        if not compatibility.is_integer(rpc_timeout):
            raise AMQPInvalidArgument('rpc_timeout should be an integer')
        elif self.is_closed:
            raise AMQPConnectionError('connection closed')

        with self.lock:
            channel_id = self._get_next_available_channel_id()
            channel = Channel(channel_id, self, rpc_timeout)
            self._channels[channel_id] = channel
            if not lazy:
                channel.open()
        LOGGER.debug('Channel #%d Opened', channel_id)
        return self._channels[channel_id]

    def check_for_errors(self):
        """Check Connection for errors.

        :raises AMQPConnectionError: Raises if the connection
                                     encountered an error.
        :return:
        """
        if not self.exceptions:
            if not self.is_closed:
                return
            why = AMQPConnectionError('connection closed')
            self.exceptions.append(why)
        self.set_state(self.CLOSED)
        self.close()
        raise self.exceptions[0]

    def close(self):
        """Close the Connection.

        :raises AMQPConnectionError: Raises if the connection
                                     encountered an error.
        :return:
        """
        LOGGER.debug('Connection Closing')
        if not self.is_closed:
            self.set_state(self.CLOSING)
        self.heartbeat.stop()
        try:
            if not self.is_closed and self.socket:
                self._channel0.send_close_connection()
                self._wait_for_connection_state(state=Stateful.CLOSED)
        except AMQPConnectionError:
            pass
        finally:
            self._close_remaining_channels()
            self._io.close()
            self.set_state(self.CLOSED)
        LOGGER.debug('Connection Closed')

    def open(self):
        """Open Connection.

        :raises AMQPConnectionError: Raises if the connection
                                     encountered an error.
        """
        LOGGER.debug('Connection Opening')
        self.set_state(self.OPENING)
        self._exceptions = []
        self._channels = {}
        self._last_channel_id = None
        self._io.open()
        self._send_handshake()
        self._wait_for_connection_state(state=Stateful.OPEN)
        self.heartbeat.start(self._exceptions)
        LOGGER.debug('Connection Opened')

    def write_frame(self, channel_id, frame_out):
        """Marshal and write an outgoing pamqp frame to the Socket.

        :param int channel_id: Channel ID.
        :param specification.Frame frame_out: Amqp frame.

        :return:
        """
        frame_data = pamqp_frame.marshal(frame_out, channel_id)
        self.heartbeat.register_write()
        self._io.write_to_socket(frame_data)

    def write_frames(self, channel_id, frames_out):
        """Marshal and write multiple outgoing pamqp frames to the Socket.

        :param int channel_id: Channel ID/
        :param list frames_out: Amqp frames.

        :return:
        """
        data_out = EMPTY_BUFFER
        for single_frame in frames_out:
            data_out += pamqp_frame.marshal(single_frame, channel_id)
        self.heartbeat.register_write()
        self._io.write_to_socket(data_out)

    def _close_remaining_channels(self):
        """Forcefully close all open channels.

        :return:
        """
        for channel_id in list(self._channels):
            self._channels[channel_id].set_state(Channel.CLOSED)
            self._channels[channel_id].close()
            self._cleanup_channel(channel_id)

    def _get_next_available_channel_id(self):
        """Returns the next available available channel id.

        :raises AMQPConnectionError: Raises if there is no available channel.
        :rtype: int
        """
        for channel_id in compatibility.RANGE(self._last_channel_id or 1,
                                              self.max_allowed_channels + 1):
            if channel_id in self._channels:
                channel = self._channels[channel_id]
                if channel.current_state != Channel.CLOSED:
                    continue
                del self._channels[channel_id]
            self._last_channel_id = channel_id
            return channel_id

        if self._last_channel_id:
            self._last_channel_id = None
            return self._get_next_available_channel_id()

        raise AMQPConnectionError(
            'reached the maximum number of channels %d' %
            self.max_allowed_channels)

    def _handle_amqp_frame(self, data_in):
        """Unmarshal a single AMQP frame and return the result.

        :param data_in: socket data

        :return: data_in, channel_id, frame
        """
        if not data_in:
            return data_in, None, None
        try:
            byte_count, channel_id, frame_in = pamqp_frame.unmarshal(data_in)
            return data_in[byte_count:], channel_id, frame_in
        except pamqp_exception.UnmarshalingException:
            pass
        except specification.AMQPFrameError as why:
            LOGGER.error('AMQPFrameError: %r', why, exc_info=True)
        except ValueError as why:
            LOGGER.error(why, exc_info=True)
            self.exceptions.append(AMQPConnectionError(why))
        return data_in, None, None

    def _read_buffer(self, data_in):
        """Process the socket buffer, and direct the data to the appropriate
        channel.

        :rtype: bytes
        """
        while data_in:
            data_in, channel_id, frame_in = self._handle_amqp_frame(data_in)

            if frame_in is None:
                break

            self.heartbeat.register_read()
            if channel_id == 0:
                self._channel0.on_frame(frame_in)
            elif channel_id in self._channels:
                self._channels[channel_id].on_frame(frame_in)

        return data_in

    def _cleanup_channel(self, channel_id):
        """Remove the the channel from the list of available channels.

        :param int channel_id: Channel id

        :return:
        """
        with self.lock:
            if channel_id not in self._channels:
                return
            del self._channels[channel_id]

    def _send_handshake(self):
        """Send a RabbitMQ Handshake.

        :return:
        """
        self._io.write_to_socket(pamqp_header.ProtocolHeader().marshal())

    def _validate_parameters(self):
        """Validate Connection Parameters.

        :return:
        """
        if not compatibility.is_string(self.parameters['hostname']):
            raise AMQPInvalidArgument('hostname should be a string')
        elif not compatibility.is_integer(self.parameters['port']):
            raise AMQPInvalidArgument('port should be an integer')
        elif not compatibility.is_string(self.parameters['username']):
            raise AMQPInvalidArgument('username should be a string')
        elif not compatibility.is_string(self.parameters['password']):
            raise AMQPInvalidArgument('password should be a string')
        elif not compatibility.is_string(self.parameters['virtual_host']):
            raise AMQPInvalidArgument('virtual_host should be a string')
        elif not isinstance(self.parameters['timeout'], (int, float)):
            raise AMQPInvalidArgument('timeout should be an integer or float')
        elif not compatibility.is_integer(self.parameters['heartbeat']):
            raise AMQPInvalidArgument('heartbeat should be an integer')

    def _wait_for_connection_state(self, state=Stateful.OPEN, rpc_timeout=30):
        """Wait for a Connection state.

        :param int state: State that we expect

        :raises AMQPConnectionError: Raises if we are unable to establish
                                     a connection to RabbitMQ.

        :return:
        """
        start_time = time.time()
        while self.current_state != state:
            self.check_for_errors()
            if time.time() - start_time > rpc_timeout:
                raise AMQPConnectionError('connection timed out')
            sleep(IDLE_WAIT)



import time

from amqpstorm.exception import AMQPChannelError
from amqpstorm.tests.functional.utility import TestFunctionalFramework
from amqpstorm.tests.functional.utility import setup


class TxFunctionalTests(TestFunctionalFramework):
    @setup()
    def test_functional_tx_select(self):
        self.channel.tx.select()

    @setup()
    def test_functional_tx_select_multiple(self):
        for _ in range(10):
            self.channel.tx.select()

    @setup(queue=True)
    def test_functional_tx_commit(self):
        self.channel.tx.select()

        self.channel.queue.declare(self.queue_name)
        self.channel.basic.publish(self.message, self.queue_name)

        # Sleep for 0.1s to make sure RabbitMQ has time to catch up.
        time.sleep(0.1)

        queue_status = self.channel.queue.declare(self.queue_name,
                                                  passive=True)
        self.assertEqual(queue_status['message_count'], 0)

        self.channel.tx.commit()

        queue_status = self.channel.queue.declare(self.queue_name,
                                                  passive=True)
        self.assertEqual(queue_status['message_count'], 1)

    @setup(queue=True)
    def test_functional_tx_commit_multiple(self):
        self.channel.tx.select()

        self.channel.queue.declare(self.queue_name)
        for _ in range(10):
            self.channel.basic.publish(self.message, self.queue_name)

        # Sleep for 0.1s to make sure RabbitMQ has time to catch up.
        time.sleep(0.1)

        queue_status = self.channel.queue.declare(self.queue_name,
                                                  passive=True)
        self.assertEqual(queue_status['message_count'], 0)

        self.channel.tx.commit()

        queue_status = self.channel.queue.declare(self.queue_name,
                                                  passive=True)
        self.assertEqual(queue_status['message_count'], 10)

    @setup()
    def test_functional_tx_commit_without_select(self):
        self.assertRaisesRegex(
            AMQPChannelError,
            'Channel 1 was closed by remote server: '
            'PRECONDITION_FAILED - channel is not transactional',
            self.channel.tx.commit
        )

    @setup(queue=True)
    def test_functional_tx_rollback(self):
        self.channel.tx.select()

        self.channel.queue.declare(self.queue_name)
        self.channel.basic.publish(self.message, self.queue_name)

        self.channel.tx.rollback()

        queue_status = self.channel.queue.declare(self.queue_name,
                                                  passive=True)
        self.assertEqual(queue_status['message_count'], 0)

    @setup(queue=True)
    def test_functional_tx_rollback_multiple(self):
        self.channel.tx.select()

        self.channel.queue.declare(self.queue_name)
        for _ in range(10):
            self.channel.basic.publish(self.message, self.queue_name)

        self.channel.tx.rollback()

        queue_status = self.channel.queue.declare(self.queue_name,
                                                  passive=True)
        self.assertEqual(queue_status['message_count'], 0)

    @setup()
    def test_functional_tx_rollback_without_select(self):
        self.assertRaisesRegex(
            AMQPChannelError,
            'Channel 1 was closed by remote server: '
            'PRECONDITION_FAILED - channel is not transactional',
            self.channel.tx.rollback
        )



from amqpstorm import Channel
from amqpstorm import exception
from amqpstorm.queue import Queue
from amqpstorm.tests.utility import FakeConnection
from amqpstorm.tests.utility import TestFramework


class QueueExceptionTests(TestFramework):
    def test_queue_declare_invalid_parameter(self):
        channel = Channel(0, FakeConnection(), 360)
        channel.set_state(Channel.OPEN)
        queue = Queue(channel)

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'queue should be a string',
            queue.declare, None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'passive should be a boolean',
            queue.declare, 'travis-ci', None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'durable should be a boolean',
            queue.declare, 'travis-ci', True, None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'exclusive should be a boolean',
            queue.declare, 'travis-ci', True, True, None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'auto_delete should be a boolean',
            queue.declare, 'travis-ci', True, True, True, None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'arguments should be a dict or None',
            queue.declare, 'travis-ci', True, True, True, True, []
        )

    def test_queue_delete_invalid_parameter(self):
        channel = Channel(0, FakeConnection(), 360)
        channel.set_state(Channel.OPEN)
        queue = Queue(channel)

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'queue should be a string',
            queue.delete, None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'if_unused should be a boolean',
            queue.delete, '', None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'if_empty should be a boolean',
            queue.delete, '', True, None
        )

    def test_queue_purge_invalid_parameter(self):
        channel = Channel(0, FakeConnection(), 360)
        channel.set_state(Channel.OPEN)
        queue = Queue(channel)

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'queue should be a string',
            queue.purge, None
        )

    def test_queue_bind_invalid_parameter(self):
        channel = Channel(0, FakeConnection(), 360)
        channel.set_state(Channel.OPEN)
        queue = Queue(channel)

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'queue should be a string',
            queue.bind, None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'exchange should be a string',
            queue.bind, '', None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'routing_key should be a string',
            queue.bind, '', '', None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'arguments should be a dict or None',
            queue.bind, '', '', '', []
        )

    def test_queue_unbind_invalid_parameter(self):
        channel = Channel(0, FakeConnection(), 360)
        channel.set_state(Channel.OPEN)
        queue = Queue(channel)

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'queue should be a string',
            queue.unbind, None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'exchange should be a string',
            queue.unbind, '', None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'routing_key should be a string',
            queue.unbind, '', '', None
        )

        self.assertRaisesRegex(
            exception.AMQPInvalidArgument,
            'arguments should be a dict or None',
            queue.unbind, '', '', '', []
        )



import time

from amqpstorm import AMQPChannelError
from amqpstorm import AMQPMessageError
from amqpstorm import Channel
from amqpstorm import Message
from amqpstorm.base import BaseMessage
from amqpstorm.tests.functional.utility import TestFunctionalFramework
from amqpstorm.tests.functional.utility import setup


class GenericTest(TestFunctionalFramework):
    def configure(self):
        self.disable_logging_validation()

    @setup(queue=True)
    def test_functional_publish_and_get_five_messages(self):
        self.channel.queue.declare(self.queue_name)

        # Publish 5 Messages.
        for _ in range(5):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        # Get 5 messages.
        for _ in range(5):
            payload = self.channel.basic.get(self.queue_name)
            self.assertIsInstance(payload, Message)
            self.assertEqual(payload.body, self.message)

    @setup(queue=True)
    def test_functional_publish_and_get_five_empty_messages(self):
        self.channel.queue.declare(self.queue_name)

        # Publish 5 Messages.
        for _ in range(5):
            self.channel.basic.publish(body=b'',
                                       routing_key=self.queue_name)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        # Get 5 messages.
        inbound_messages = []
        for _ in range(5):
            payload = self.channel.basic.get(self.queue_name)
            self.assertIsInstance(payload, Message)
            self.assertEqual(payload.body, b'')
            inbound_messages.append(payload)

        self.assertEqual(len(inbound_messages), 5)

    @setup(queue=True)
    def test_functional_publish_and_get_a_large_message(self):
        self.channel.confirm_deliveries()
        self.channel.queue.declare(self.queue_name)

        body = self.message * 65536

        # Publish a single large message
        self.channel.basic.publish(body=body,
                                   routing_key=self.queue_name)

        payload = self.channel.basic.get(self.queue_name)
        self.assertEqual(body, payload.body)

    @setup(queue=True)
    def test_functional_publish_5_large_messages_and_consume(self):
        self.channel.confirm_deliveries()
        self.channel.queue.declare(self.queue_name)

        body = self.message * 8192
        messages_to_publish = 5

        self.channel.basic.consume(queue=self.queue_name,
                                   no_ack=True)
        # Publish 5 Messages.
        for _ in range(messages_to_publish):
            self.channel.basic.publish(body=body,
                                       routing_key=self.queue_name)

        inbound_messages = []
        for message in self.channel.build_inbound_messages(
                break_on_empty=True):
            self.assertEqual(message.body, body)
            inbound_messages.append(message)
        self.assertEqual(len(inbound_messages), messages_to_publish)

    @setup(queue=True)
    def test_functional_publish_5_empty_messages(self):
        self.channel.confirm_deliveries()
        self.channel.queue.declare(self.queue_name)

        body = b''
        messages_to_publish = 5

        self.channel.basic.consume(queue=self.queue_name,
                                   no_ack=True)
        # Publish 5 Messages.
        for _ in range(messages_to_publish):
            self.channel.basic.publish(body=body,
                                       routing_key=self.queue_name)

        inbound_messages = []
        for message in self.channel.build_inbound_messages(
                break_on_empty=True):
            self.assertEqual(message.body, body)
            inbound_messages.append(message)
        self.assertEqual(len(inbound_messages), messages_to_publish)

    @setup(queue=True)
    def test_functional_publish_5_large_messages_and_get(self):
        self.channel.confirm_deliveries()
        self.channel.queue.declare(self.queue_name)

        body = self.message * 8192
        messages_to_publish = 5

        # Publish 5 Messages.
        for _ in range(messages_to_publish):
            self.channel.basic.publish(body=body,
                                       routing_key=self.queue_name)

        inbound_messages = []
        for _ in range(messages_to_publish):
            message = self.channel.basic.get(self.queue_name,
                                             no_ack=True)
            self.assertEqual(message.body, body)
            inbound_messages.append(message)
        self.assertEqual(len(inbound_messages), messages_to_publish)

    @setup(queue=True)
    def test_functional_publish_with_properties_and_get(self):
        self.channel.confirm_deliveries()
        self.channel.queue.declare(self.queue_name)

        app_id = 'travis-ci'
        properties = {
            'headers': {
                'key': 1234567890,
                'alpha': 'omega'
            }
        }

        message = Message.create(self.channel,
                                 body=self.message,
                                 properties=properties)
        # Assign Property app_id
        message.app_id = app_id

        # Check that it was set correctly.
        self.assertEqual(message.properties['app_id'], app_id)

        # Get Property Correlation Id
        correlation_id = message.correlation_id

        # Publish Message
        message.publish(routing_key=self.queue_name)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        # New way
        payload = self.channel.basic.get(self.queue_name)
        self.assertEqual(payload.properties['headers']['key'], 1234567890)
        self.assertEqual(payload.properties['headers']['alpha'], 'omega')
        self.assertEqual(payload.app_id, app_id)
        self.assertEqual(payload.correlation_id, correlation_id)
        self.assertIsInstance(payload.properties['app_id'], str)
        self.assertIsInstance(payload.properties['correlation_id'], str)

        # Old way
        result = payload.to_dict()
        self.assertEqual(result['properties']['headers']['key'], 1234567890)
        self.assertEqual(result['properties']['headers']['alpha'], b'omega')
        self.assertIsInstance(result['properties']['app_id'], str)
        self.assertIsInstance(result['properties']['correlation_id'], str)
        self.assertEqual(result['properties']['app_id'], app_id)
        self.assertEqual(result['properties']['correlation_id'],
                         correlation_id)

    @setup(queue=True)
    def test_functional_publish_and_change_app_id(self):
        self.channel.confirm_deliveries()
        self.channel.queue.declare(self.queue_name)
        message = Message.create(self.channel,
                                 body=self.message)
        message.app_id = 'travis-ci'
        message.publish(self.queue_name)

        message = self.channel.basic.get(self.queue_name, no_ack=True)

        # Check original app_id
        self.assertEqual(message.app_id, 'travis-ci')

        # Assign Property app_id
        app_id = 'travis-ci-2'.encode('utf-8')
        message.app_id = app_id

        # Check that it was set correctly.
        self.assertEqual(message.properties['app_id'], app_id)

        # Get Property Correlation Id
        correlation_id = message.correlation_id

        # Publish Message
        message.publish(routing_key=self.queue_name)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        payload = self.channel.basic.get(self.queue_name, no_ack=True)
        self.assertEqual(payload.app_id, app_id.decode('utf-8'))
        self.assertEqual(payload.correlation_id, correlation_id)
        self.assertIsInstance(payload.properties['app_id'], str)
        self.assertIsInstance(payload.properties['correlation_id'], str)

    @setup(queue=True)
    def test_functional_publish_and_consume_five_messages(self):
        self.channel.confirm_deliveries()
        self.channel.queue.declare(self.queue_name)

        for _ in range(5):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)

        # Store and inbound messages.
        inbound_messages = []

        def on_message(message):
            self.assertIsInstance(message.body, (bytes, str))
            self.assertIsInstance(message.channel, Channel)
            self.assertIsInstance(message.properties, dict)
            self.assertIsInstance(message.method, dict)
            inbound_messages.append(message)

        self.channel.basic.consume(callback=on_message,
                                   queue=self.queue_name,
                                   no_ack=True)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        self.channel.process_data_events()

        # Make sure all five messages were downloaded.
        self.assertEqual(len(inbound_messages), 5)

    @setup(queue=True)
    def test_functional_generator_consume(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        for _ in range(5):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)
        self.channel.basic.consume(queue=self.queue_name,
                                   no_ack=True)
        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        # Store and inbound messages.
        inbound_messages = []

        for message in self.channel.build_inbound_messages(
                break_on_empty=True):
            self.assertIsInstance(message, Message)
            inbound_messages.append(message)

        # Make sure all five messages were downloaded.
        self.assertEqual(len(inbound_messages), 5)

    @setup(queue=True)
    def test_functional_consume_with_custom_message_impl(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        for _ in range(5):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)
        self.channel.basic.consume(queue=self.queue_name,
                                   no_ack=True)
        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        # Store and inbound messages.
        inbound_messages = []

        class CustomMessage(BaseMessage):
            pass

        for message in self.channel.build_inbound_messages(
                break_on_empty=True,
                message_impl=CustomMessage):
            self.assertIsInstance(message, CustomMessage)
            self.assertEqual(message._body, self.message.encode('utf-8'))
            inbound_messages.append(message)

        # Make sure all five messages were downloaded.
        self.assertEqual(len(inbound_messages), 5)

    @setup(queue=True)
    def test_functional_consume_and_redeliver(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        self.channel.basic.publish(body=self.message,
                                   routing_key=self.queue_name)

        def on_message_first(message):
            self.channel.stop_consuming()
            message.reject()

        self.channel.basic.consume(callback=on_message_first,
                                   queue=self.queue_name,
                                   no_ack=False)
        self.channel.process_data_events()

        # Store and inbound messages.
        inbound_messages = []

        # Close current channel and open a new one.
        self.channel.close()

        # Sleep for 0.1s to make sure RabbitMQ has time to catch up.
        time.sleep(0.1)

        channel = self.connection.channel()

        def on_message_second(message):
            inbound_messages.append(message)
            self.assertEqual(message.body, self.message)

        channel.basic.consume(callback=on_message_second,
                              queue=self.queue_name,
                              no_ack=True)
        channel.process_data_events()

        # Sleep for 0.1s to make sure RabbitMQ has time to catch up.
        time.sleep(0.1)

        start_time = time.time()
        while len(inbound_messages) != 1:
            if time.time() - start_time >= 30:
                break
            time.sleep(0.1)

        self.assertEqual(len(inbound_messages), 1)

    @setup(queue=True)
    def test_functional_redelivered(self):
        self.channel.queue.declare(self.queue_name)

        self.channel.confirm_deliveries()
        self.channel.basic.publish(body=self.message,
                                   routing_key=self.queue_name,
                                   mandatory=True)

        # Sleep for 0.1s to make sure RabbitMQ has time to catch up.
        time.sleep(0.1)

        message = self.channel.basic.get(self.queue_name, no_ack=False)
        message.reject(requeue=True)

        # Store and inbound messages.
        inbound_messages = []

        def on_message(message):
            inbound_messages.append(message)
            self.assertTrue(message.redelivered)

        self.channel.basic.consume(callback=on_message,
                                   queue=self.queue_name,
                                   no_ack=True)

        start_time = time.time()
        while len(inbound_messages) == 0:
            self.channel.process_data_events()
            if time.time() - start_time >= 30:
                break
            time.sleep(0.1)

        self.assertEqual(len(inbound_messages), 1)

    @setup(queue=True)
    def test_functional_get_and_redeliver(self):
        self.channel.queue.declare(self.queue_name)

        self.channel.confirm_deliveries()
        self.channel.basic.publish(body=self.message,
                                   routing_key=self.queue_name)

        # Sleep for 0.1s to make sure RabbitMQ has time to catch up.
        time.sleep(0.1)

        message = self.channel.basic.get(self.queue_name, no_ack=False)
        message.reject(requeue=True)

        # Sleep for 0.1s to make sure RabbitMQ has time to catch up.
        time.sleep(0.1)

        message = self.channel.basic.get(self.queue_name, no_ack=True)
        self.assertEqual(message.body, self.message)

    @setup(queue=True)
    def test_functional_publish_and_get(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        self.channel.basic.publish(body=self.message,
                                   routing_key=self.queue_name)

        message = self.channel.basic.get(self.queue_name, no_ack=True,
                                         auto_decode=False)
        self.assertIsInstance(message.body, bytes)
        self.assertEqual(message.body, self.message.encode('utf-8'))

    @setup(queue=True)
    def test_functional_publish_and_get_with_custom_message_impl(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        self.channel.basic.publish(body=self.message,
                                   routing_key=self.queue_name)

        class CustomMessage(BaseMessage):
            pass

        message = self.channel.basic.get(self.queue_name, no_ack=True,
                                         auto_decode=False,
                                         message_impl=CustomMessage)
        self.assertIsInstance(message, CustomMessage)
        self.assertEqual(message._body, self.message.encode('utf-8'))

    @setup(queue=True)
    def test_functional_publish_and_get_auto_decode(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        self.channel.basic.publish(body=self.message,
                                   routing_key=self.queue_name)

        message = self.channel.basic.get(self.queue_name, no_ack=False,
                                         auto_decode=True)
        self.assertIsInstance(message.body, str)
        self.assertEqual(message.body, self.message)

    @setup(queue=True)
    def test_functional_publish_and_consume_auto_decode(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        self.channel.basic.publish(body=self.message,
                                   routing_key=self.queue_name)

        def on_message(message):
            self.assertIsInstance(message.body, str)
            self.assertEqual(message.body, self.message)
            message.channel.stop_consuming()

        self.channel.basic.consume(callback=on_message,
                                   queue=self.queue_name,
                                   no_ack=True)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        self.channel.start_consuming(auto_decode=True)

    @setup(queue=True)
    def test_functional_publish_and_consume(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        self.channel.basic.publish(body=self.message,
                                   routing_key=self.queue_name)

        def on_message(message):
            self.assertIsInstance(message.body, bytes)
            self.assertEqual(message.body, self.message.encode('utf-8'))
            message.channel.stop_consuming()

        self.channel.basic.consume(callback=on_message,
                                   queue=self.queue_name,
                                   no_ack=True)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        self.channel.start_consuming(auto_decode=False)

    @setup(queue=True)
    def test_functional_publish_and_confirm(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()

        self.channel.basic.publish(body=self.message,
                                   routing_key=self.queue_name)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        payload = self.channel.queue.declare(self.queue_name,
                                             passive=True)
        self.assertEqual(payload['message_count'], 1)

    @setup(queue=True)
    def test_functional_publish_confirm_to_invalid_queue(self):
        self.channel.confirm_deliveries()
        self.assertRaises(AMQPMessageError,
                          self.channel.basic.publish,
                          body=self.message,
                          exchange='amq.direct',
                          mandatory=True,
                          routing_key=self.queue_name)

    @setup(queue=True)
    def test_functional_publish_fail_publish_and_confirm(self):
        self.channel.confirm_deliveries()
        try:
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name,
                                       mandatory=True)
        except AMQPChannelError as why:
            self.assertTrue(self.channel.is_open)
            self.assertEqual(why.error_code, 312)
            if why.error_code == 312:
                self.channel.queue.declare(self.queue_name)

        result = self.channel.basic.publish(
            body=self.message,
            routing_key=self.queue_name,
            mandatory=True
        )
        self.assertTrue(result)

        payload = self.channel.queue.declare(self.queue_name,
                                             passive=True)
        self.assertEqual(payload['message_count'], 1)

    @setup(queue=True)
    def test_functional_publish_fail_recover(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()

        message = Message.create(self.channel, self.message)
        self.assertRaises(AMQPChannelError, message.publish, self.message,
                          exchange=self.exchange_name, mandatory=True)
        self.assertFalse(self.channel.is_open)

    @setup(queue=True)
    def test_functional_start_stop_consumer(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()

        for _ in range(5):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)

        # Store and inbound messages.
        inbound_messages = []

        def on_message(message):
            self.assertIsInstance(message.body, (bytes, str))
            self.assertIsInstance(message.channel, Channel)
            self.assertIsInstance(message.properties, dict)
            self.assertIsInstance(message.method, dict)
            inbound_messages.append(message)
            if len(inbound_messages) >= 5:
                message.channel.stop_consuming()

        self.channel.basic.consume(callback=on_message,
                                   queue=self.queue_name,
                                   no_ack=True)

        # Sleep for 0.01s to make sure RabbitMQ has time to catch up.
        time.sleep(0.01)

        self.channel.start_consuming()

        # Make sure all five messages were downloaded.
        self.assertEqual(len(inbound_messages), 5)

    @setup(new_channel=False)
    def test_functional_reuse_channel_with_confirm_deliveries(self):
        channel = self.connection.channel()
        channel.confirm_deliveries()
        self.assertTrue(channel._confirming_deliveries)
        self.assertTrue(channel.basic.publish(body=self.message,
                                              routing_key=self.queue_name))

        channel.close()
        channel.open()

        # Make sure the re-opened channel does not have confirm deliveries
        # enabled.
        self.assertFalse(channel._confirming_deliveries)
        self.assertIsNone(channel.basic.publish(body=self.message,
                                                routing_key=self.queue_name))

        # But make sure we can safely enable it again.
        channel.confirm_deliveries()
        self.assertTrue(channel._confirming_deliveries)
        self.assertTrue(channel.basic.publish(body=self.message,
                                              routing_key=self.queue_name))
        channel.close()



import importlib
import ssl
import sys

from amqpstorm import AMQPConnectionError
from amqpstorm import UriConnection
from amqpstorm import compatibility
from amqpstorm.tests.utility import TestFramework
from amqpstorm.tests.utility import unittest


class UriConnectionExceptionTests(TestFramework):
    @unittest.skipIf(sys.version_info < (3, 3), 'Python 3.x test')
    def test_uri_py3_raises_on_invalid_uri(self):
        self.assertRaises(ValueError, UriConnection, 'amqp://a:b', {}, True)

    @unittest.skipIf(sys.version_info[0] == 3, 'Python 2.x test')
    def test_uri_py2_raises_on_invalid_uri(self):
        self.assertRaises(ValueError, UriConnection, 'amqp://a:b', {}, True)

    def test_uri_raises_on_invalid_object(self):
        self.assertRaises(AttributeError, UriConnection, None)
        self.assertRaises(AttributeError, UriConnection, {})
        self.assertRaises(AttributeError, UriConnection, [])
        self.assertRaises(AttributeError, UriConnection, ())

    def test_uri_invalid_ssl_options(self):
        connection = UriConnection(
            'amqps://guest:guest@localhost:5672/%2F', lazy=True
        )
        ssl_kwargs = {
            'unit_test': ['not_required'],
        }
        ssl_options = connection._parse_ssl_options(ssl_kwargs)

        self.assertFalse(ssl_options)
        self.assertIn("invalid option: unit_test",
                      self.get_last_log())

    def test_uri_get_invalid_ssl_version(self):
        connection = UriConnection(
            'amqps://guest:guest@localhost:5672/%2F', lazy=True
        )

        self.assertEqual(connection._get_ssl_version('protocol_test'),
                         ssl.PROTOCOL_TLSv1)
        self.assertIn("ssl_options: ssl_version 'protocol_test' not found "
                      "falling back to PROTOCOL_TLSv1.",
                      self.get_last_log())

    def test_uri_get_invalid_ssl_validation(self):
        connection = UriConnection(
            'amqps://guest:guest@localhost:5672/%2F', lazy=True
        )

        self.assertEqual(ssl.CERT_NONE,
                         connection._get_ssl_validation('cert_test'))
        self.assertIn("ssl_options: cert_reqs 'cert_test' not found "
                      "falling back to CERT_NONE.",
                      self.get_last_log())

    def test_uri_ssl_not_supported(self):
        restore_func = sys.modules['ssl']
        try:
            sys.modules['ssl'] = None
            importlib.reload(compatibility)
            self.assertIsNone(compatibility.ssl)
            self.assertRaisesRegex(
                AMQPConnectionError,
                'Python not compiled with support for TLSv1 or higher',
                UriConnection, 'amqps://localhost:5672/%2F'
            )
        finally:
            sys.modules['ssl'] = restore_func
            importlib.reload(compatibility)



from amqpstorm.tests.functional.utility import TestFunctionalFramework
from amqpstorm.tests.functional.utility import setup


class BasicFunctionalTests(TestFunctionalFramework):
    @setup()
    def test_functional_basic_qos(self):
        result = self.channel.basic.qos(prefetch_count=100)
        self.assertFalse(result)

    @setup(queue=True)
    def test_functional_basic_get(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.basic.publish(self.message, self.queue_name)

        message = self.channel.basic.get(self.queue_name)
        self.assertEqual(message.body, self.message)
        message.ack()

    @setup(queue=True)
    def test_functional_basic_get_empty(self):
        self.channel.queue.declare(self.queue_name)

        message = self.channel.basic.get(self.queue_name)
        self.assertIsNone(message)

    @setup(queue=True)
    def test_functional_basic_cancel(self):
        self.channel.queue.declare(self.queue_name)
        consumer_tag = self.channel.basic.consume(None, self.queue_name)

        result = self.channel.basic.cancel(consumer_tag)
        self.assertEqual(result['consumer_tag'], consumer_tag)

    @setup(queue=True)
    def test_functional_basic_recover(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.basic.publish(self.message, self.queue_name)

        self.assertFalse(self.channel.basic.recover(requeue=True))

    @setup(queue=True)
    def test_functional_basic_ack(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.basic.publish(self.message, self.queue_name)

        message = self.channel.basic.get(self.queue_name)
        result = self.channel.basic.ack(delivery_tag=message.delivery_tag)

        self.assertIsNone(result)

        # Make sure the message wasn't requeued.
        self.assertFalse(self.channel.basic.get(self.queue_name))

    @setup(queue=True)
    def test_functional_basic_ack_multiple(self):
        message = None
        self.channel.queue.declare(self.queue_name)

        for _ in range(5):
            self.channel.basic.publish(self.message, self.queue_name)

        for _ in range(5):
            message = self.channel.basic.get(self.queue_name)

        self.assertIsNotNone(message)

        self.channel.basic.ack(
            delivery_tag=message.delivery_tag,
            multiple=True
        )

        # Make sure the message wasn't requeued.
        self.assertFalse(self.channel.basic.get(self.queue_name))

    @setup(queue=True)
    def test_functional_basic_nack(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.basic.publish(self.message, self.queue_name)

        message = self.channel.basic.get(self.queue_name)
        result = self.channel.basic.nack(
            requeue=False,
            delivery_tag=message.delivery_tag
        )

        self.assertIsNone(result)

        # Make sure the message wasn't requeued.
        self.assertFalse(self.channel.basic.get(self.queue_name))

    @setup(queue=True)
    def test_functional_basic_nack_multiple(self):
        message = None
        self.channel.queue.declare(self.queue_name)

        for _ in range(5):
            self.channel.basic.publish(self.message, self.queue_name)

        for _ in range(5):
            message = self.channel.basic.get(self.queue_name)

        self.assertIsNotNone(message)
        self.channel.basic.nack(
            delivery_tag=message.delivery_tag,
            requeue=False,
            multiple=True
        )

        # Make sure the message wasn't requeued.
        self.assertFalse(self.channel.basic.get(self.queue_name))

    @setup(queue=True)
    def test_functional_basic_nack_requeue(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.basic.publish(self.message, self.queue_name)

        message = self.channel.basic.get(self.queue_name)
        result = self.channel.basic.nack(
            requeue=True,
            delivery_tag=message.delivery_tag
        )

        self.assertIsNone(result)

        # Make sure the message was requeued.
        self.assertIsNotNone(self.channel.basic.get(self.queue_name))

    @setup(queue=True)
    def test_functional_basic_reject(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.basic.publish(self.message, self.queue_name)

        message = self.channel.basic.get(self.queue_name)
        result = self.channel.basic.reject(
            requeue=False,
            delivery_tag=message.delivery_tag
        )

        self.assertIsNone(result)

        # Make sure the message wasn't requeued.
        self.assertFalse(self.channel.basic.get(self.queue_name))

    @setup(queue=True)
    def test_functional_basic_reject_requeue(self):
        self.channel.queue.declare(self.queue_name)
        self.channel.basic.publish(self.message, self.queue_name)

        message = self.channel.basic.get(self.queue_name)
        result = self.channel.basic.reject(
            requeue=True,
            delivery_tag=message.delivery_tag
        )

        self.assertIsNone(result)

        # Make sure the message was requeued.
        self.assertIsNotNone(self.channel.basic.get(self.queue_name))



from amqpstorm.management import ManagementApi
from amqpstorm.tests import HTTP_URL
from amqpstorm.tests import PASSWORD
from amqpstorm.tests import USERNAME
from amqpstorm.tests.functional.utility import TestFunctionalFramework


class ApiUserFunctionalTests(TestFunctionalFramework):
    def test_api_virtual_host_get(self):
        api = ManagementApi(HTTP_URL, USERNAME, PASSWORD)

        virtual_host = api.virtual_host.get('/')
        self.assertIsInstance(virtual_host, dict)
        self.assertEqual(virtual_host['name'], '/')

    def test_api_virtual_host_list(self):
        api = ManagementApi(HTTP_URL, USERNAME, PASSWORD)

        virtual_hosts = api.virtual_host.list()
        self.assertIsInstance(virtual_hosts, list)
        self.assertGreater(len(virtual_hosts), 0)

        for vhost in virtual_hosts:
            self.assertIsInstance(vhost, dict)
            self.assertIn('name', vhost)

    def test_api_virtual_host_create(self):
        vhost_name = 'test_api_virtual_host_create'
        api = ManagementApi(HTTP_URL, USERNAME, PASSWORD)

        try:
            self.assertIsNone(api.virtual_host.create(vhost_name))
            virtual_host = api.virtual_host.get(vhost_name)
            self.assertIsInstance(virtual_host, dict)
            self.assertEqual(virtual_host['name'], vhost_name)

            api.user.set_permission(USERNAME, vhost_name)
        finally:
            api.virtual_host.delete(vhost_name)

    def test_api_virtual_host_permissions(self):
        api = ManagementApi(HTTP_URL, USERNAME, PASSWORD)

        permissions = api.virtual_host.get_permissions('/')
        self.assertIsInstance(permissions, list)
        self.assertIsInstance(permissions[0], dict)
        self.assertEqual(permissions[0]['read'], '.*')
        self.assertEqual(permissions[0]['write'], '.*')
        self.assertEqual(permissions[0]['configure'], '.*')
        self.assertEqual(permissions[0]['user'], USERNAME)
        self.assertEqual(permissions[0]['vhost'], '/')



import os.path
import ssl
import threading
import time
import unittest

from amqpstorm import Connection
from amqpstorm import UriConnection
from amqpstorm.tests import CAFILE
from amqpstorm.tests import CERTFILE
from amqpstorm.tests import KEYFILE
from amqpstorm.tests import PASSWORD
from amqpstorm.tests import SSL_HOST
from amqpstorm.tests import SSL_URI
from amqpstorm.tests import USERNAME
from amqpstorm.tests.functional.utility import TestFunctionalFramework
from amqpstorm.tests.functional.utility import setup


@unittest.skipIf(not os.path.exists(CAFILE), 'SSL tests not configured')
class SSLReliabilityFunctionalTests(TestFunctionalFramework):
    @setup(new_connection=False, queue=True)
    def test_functional_ssl_open_new_connection_loop(self):
        context = ssl.create_default_context(cafile=CAFILE)
        context.load_cert_chain(
            certfile=CERTFILE,
            keyfile=KEYFILE,
        )
        ssl_options = {
            'context': context,
            'server_hostname': SSL_HOST
        }

        for _ in range(5):
            self.connection = self.connection = Connection(
                SSL_HOST, USERNAME, PASSWORD, port=5671, ssl=True,
                ssl_options=ssl_options, timeout=1)
            self.channel = self.connection.channel()

            # Make sure that it's a new channel.
            self.assertEqual(int(self.channel), 1)

            self.channel.queue.declare(self.queue_name)

            # Verify that the Connection/Channel has been opened properly.
            self.assertIsNotNone(self.connection._io.socket)
            self.assertIsNotNone(self.connection._io.poller)
            self.assertTrue(self.connection.is_open)

            self.channel.close()
            self.connection.close()

            # Verify that the Connection has been closed properly.
            self.assertTrue(self.connection.is_closed)
            self.assertIsNone(self.connection._io.socket)
            self.assertIsNone(self.connection._io.poller)
            self.assertFalse(self.connection._io._running.is_set())
            self.assertFalse(self.connection.exceptions)

    @setup(new_connection=False, queue=True)
    def test_functional_ssl_open_close_connection_loop(self):
        context = ssl.create_default_context(cafile=CAFILE)
        context.load_cert_chain(
            certfile=CERTFILE,
            keyfile=KEYFILE,
        )
        ssl_options = {
            'context': context,
            'server_hostname': SSL_HOST
        }
        self.connection = self.connection = Connection(
            SSL_HOST, USERNAME, PASSWORD, port=5671, ssl=True,
            ssl_options=ssl_options, timeout=1, lazy=True)

        for _ in range(5):
            self.connection.open()
            channel = self.connection.channel()

            # Make sure that it's a new channel.
            self.assertEqual(int(channel), 1)

            channel.queue.declare(self.queue_name)

            channel.close()

            # Verify that the Connection/Channel has been opened properly.
            self.assertIsNotNone(self.connection._io.socket)
            self.assertIsNotNone(self.connection._io.poller)
            self.assertTrue(self.connection.is_open)

            self.connection.close()

            # Verify that the Connection has been closed properly.
            self.assertTrue(self.connection.is_closed)
            self.assertIsNone(self.connection._io.socket)
            self.assertIsNone(self.connection._io.poller)
            self.assertFalse(self.connection._io._running.is_set())
            self.assertFalse(self.connection.exceptions)

    @setup(new_connection=False, queue=False)
    def test_functional_ssl_open_close_channel_loop(self):
        context = ssl.create_default_context(cafile=CAFILE)
        context.load_cert_chain(
            certfile=CERTFILE,
            keyfile=KEYFILE,
        )
        ssl_options = {
            'context': context,
            'server_hostname': SSL_HOST
        }
        self.connection = Connection(
            SSL_HOST, USERNAME, PASSWORD, port=5671, ssl=True,
            ssl_options=ssl_options)

        for _ in range(25):
            channel = self.connection.channel()

            # Verify that the Channel has been opened properly.
            self.assertTrue(self.connection.is_open)
            self.assertTrue(channel.is_open)

            # Channel id should be staying at 1.
            self.assertEqual(int(channel), 1)

            channel.close()

            # Verify that theChannel has been closed properly.
            self.assertTrue(self.connection.is_open)
            self.assertTrue(channel.is_closed)

    @setup(new_connection=False, queue=True)
    def test_functional_ssl_open_multiple_channels(self):
        context = ssl.create_default_context(cafile=CAFILE)
        context.load_cert_chain(
            certfile=CERTFILE,
            keyfile=KEYFILE,
        )
        ssl_options = {
            'context': context,
            'server_hostname': SSL_HOST
        }
        self.connection = self.connection = Connection(
            SSL_HOST, USERNAME, PASSWORD, port=5671, ssl=True,
            ssl_options=ssl_options, lazy=True)

        for _ in range(5):
            channels = []
            self.connection.open()
            for index in range(3):
                channel = self.connection.channel()
                channels.append(channel)

                # Verify that the Channel has been opened properly.
                self.assertTrue(channel.is_open)
                self.assertEqual(int(channel), len(channels))
            self.connection.close()

    @setup(new_connection=False, queue=False)
    def test_functional_ssl_close_performance(self):
        """Make sure closing a connection never takes longer than ~1 seconds.

        :return:
        """
        for _ in range(10):
            context = ssl.create_default_context(cafile=CAFILE)
            context.load_cert_chain(
                certfile=CERTFILE,
                keyfile=KEYFILE,
            )
            ssl_options = {
                'context': context,
                'server_hostname': SSL_HOST
            }
            self.connection = self.connection = Connection(
                SSL_HOST, USERNAME, PASSWORD, timeout=60, port=5671, ssl=True,
                ssl_options=ssl_options)

            start_time = time.time()
            self.connection.close()
            self.assertLess(time.time() - start_time, 3)

    @setup(new_connection=False)
    def test_functional_ssl_uri_connection(self):
        context = ssl.create_default_context(cafile=CAFILE)
        context.load_cert_chain(
            certfile=CERTFILE,
            keyfile=KEYFILE,
        )
        ssl_options = {
            'context': context,
            'server_hostname': SSL_HOST
        }

        self.connection = UriConnection(SSL_URI, ssl_options=ssl_options)
        self.channel = self.connection.channel()
        self.assertTrue(self.connection.is_open)

    @setup(new_connection=False)
    def test_functional_ssl_uri_connection_with_context(self):
        context = ssl.create_default_context(cafile=CAFILE)
        context.load_cert_chain(
            certfile=CERTFILE,
            keyfile=KEYFILE,
        )
        ssl_options = {
            'context': context,
            'server_hostname': SSL_HOST
        }

        self.connection = UriConnection(SSL_URI, ssl_options=ssl_options)
        self.channel = self.connection.channel()
        self.assertTrue(self.connection.is_open)


@unittest.skipIf(not os.path.exists(CAFILE), 'SSL tests not configured')
class PublishAndConsume1kWithSSLTest(TestFunctionalFramework):
    messages_to_send = 1000
    messages_consumed = 0
    lock = threading.Lock()

    def configure(self):
        self.disable_logging_validation()

    def publish_messages(self):
        for _ in range(self.messages_to_send):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)

    def consume_messages(self):
        channel = self.connection.channel()
        channel.basic.consume(queue=self.queue_name,
                              no_ack=False)
        for message in channel.build_inbound_messages(
                break_on_empty=False):
            self.increment_message_count()
            message.ack()
            if self.messages_consumed == self.messages_to_send:
                break

    def increment_message_count(self):
        with self.lock:
            self.messages_consumed += 1

    @setup(new_connection=False, queue=False)
    def test_functional_publish_1k_with_ssl(self):
        context = ssl.create_default_context(cafile=CAFILE)
        context.load_cert_chain(
            certfile=CERTFILE,
            keyfile=KEYFILE,
        )
        ssl_options = {
            'context': context,
            'server_hostname': SSL_HOST
        }
        self.connection = self.connection = Connection(
            SSL_HOST, USERNAME, PASSWORD, port=5671, ssl=True,
            ssl_options=ssl_options)

        self.channel = self.connection.channel()
        self.channel.queue.declare(self.queue_name)

        publish_thread = threading.Thread(target=self.publish_messages, )
        publish_thread.daemon = True
        publish_thread.start()

        for _ in range(4):
            consumer_thread = threading.Thread(target=self.consume_messages, )
            consumer_thread.daemon = True
            consumer_thread.start()

        start_time = time.time()
        while self.messages_consumed != self.messages_to_send:
            if time.time() - start_time >= 60:
                break
            time.sleep(0.1)

        for channel in list(self.connection.channels.values()):
            channel.stop_consuming()
            channel.close()

        self.assertEqual(self.messages_consumed, self.messages_to_send,
                         'test took too long')


@unittest.skipIf(not os.path.exists(CAFILE), 'SSL tests not configured')
class Consume1kWithSSLUntilEmpty(TestFunctionalFramework):
    messages_to_send = 1000

    def configure(self):
        self.disable_logging_validation()

    def publish_messages(self):
        for _ in range(self.messages_to_send):
            self.channel.basic.publish(body=self.message,
                                       routing_key=self.queue_name)

    @setup(new_connection=False, queue=True)
    def test_functional_consume_with_ssl_until_empty(self):
        context = ssl.create_default_context(cafile=CAFILE)
        context.load_cert_chain(
            certfile=CERTFILE,
            keyfile=KEYFILE,
        )
        ssl_options = {
            'context': context,
            'server_hostname': SSL_HOST
        }
        self.connection = self.connection = Connection(
            SSL_HOST, USERNAME, PASSWORD, port=5671, ssl=True,
            ssl_options=ssl_options)

        self.channel = self.connection.channel()
        self.channel.queue.declare(self.queue_name)
        self.channel.confirm_deliveries()
        self.publish_messages()

        channel = self.connection.channel()
        channel.basic.consume(queue=self.queue_name,
                              no_ack=False)
        message_count = 0
        for message in channel.build_inbound_messages(break_on_empty=True):
            message_count += 1
            message.ack()

        result = channel.queue.declare(self.queue_name, passive=True)
        self.assertEqual(result['message_count'], 0)
        self.assertEqual(message_count, self.messages_to_send,
                         'not all messages consumed')

        channel.close()



"""AMQPStorm Connection.IO."""

import logging
import select
import socket
import threading
from errno import EAGAIN
from errno import EINTR
from errno import EWOULDBLOCK

from amqpstorm import compatibility
from amqpstorm.base import MAX_FRAME_SIZE
from amqpstorm.compatibility import ssl
from amqpstorm.exception import AMQPConnectionError

EMPTY_BUFFER = bytes()
LOGGER = logging.getLogger(__name__)
POLL_TIMEOUT = 1.0


class Poller(object):
    """Socket Read Poller."""

    def __init__(self, fileno, exceptions, timeout=5):
        self.select = select
        self._fileno = fileno
        self._exceptions = exceptions
        self.timeout = timeout

    @property
    def fileno(self):
        """Socket Fileno.

        :return:
        """
        return self._fileno

    @property
    def is_ready(self):
        """Is Socket Ready.

        :rtype: tuple
        """
        try:
            ready, _, _ = self.select.select([self.fileno], [], [],
                                             POLL_TIMEOUT)
            return bool(ready)
        except self.select.error as why:
            if why.args[0] != EINTR:
                self._exceptions.append(AMQPConnectionError(why))
        return False


class IO(object):
    """Internal Input/Output handler."""

    def __init__(self, parameters, exceptions=None, on_read_impl=None):
        self._exceptions = exceptions
        self._wr_lock = threading.Lock()
        self._rd_lock = threading.Lock()
        self._inbound_thread = None
        self._on_read_impl = on_read_impl
        self._running = threading.Event()
        self._parameters = parameters
        self.data_in = EMPTY_BUFFER
        self.poller = None
        self.socket = None
        self.use_ssl = self._parameters['ssl']

    def close(self):
        """Close Socket.

        :return:
        """
        self._wr_lock.acquire()
        self._rd_lock.acquire()
        try:
            self._running.clear()
            self._close_socket()
        finally:
            self._wr_lock.release()
            self._rd_lock.release()

        if self._inbound_thread:
            self._inbound_thread.join(timeout=self._parameters['timeout'])

        self.socket = None
        self.poller = None
        self._inbound_thread = None

    def open(self):
        """Open Socket and establish a connection.

        :raises AMQPConnectionError: Raises if the connection
                                     encountered an error.
        :return:
        """
        self._wr_lock.acquire()
        self._rd_lock.acquire()
        try:
            self.data_in = EMPTY_BUFFER
            self._running.set()
            sock_addresses = self._get_socket_addresses()
            self.socket = self._find_address_and_connect(sock_addresses)
            self.poller = Poller(self.socket.fileno(), self._exceptions,
                                 timeout=self._parameters['timeout'])
            self._inbound_thread = self._create_inbound_thread()
        finally:
            self._wr_lock.release()
            self._rd_lock.release()

    def write_to_socket(self, frame_data):
        """Write data to the socket.

        :param str frame_data:
        :return:
        """
        self._wr_lock.acquire()
        try:
            total_bytes_written = 0
            bytes_to_send = len(frame_data)
            while total_bytes_written < bytes_to_send:
                try:
                    if not self.socket:
                        raise socket.error('connection/socket error')
                    bytes_written = (
                        self.socket.send(frame_data[total_bytes_written:])
                    )
                    if bytes_written == 0:
                        raise socket.error('connection/socket error')
                    total_bytes_written += bytes_written
                except socket.timeout:
                    pass
                except socket.error as why:
                    if why.args[0] in (EWOULDBLOCK, EAGAIN):
                        continue
                    self._exceptions.append(AMQPConnectionError(why))
                    return
        finally:
            self._wr_lock.release()

    def _close_socket(self):
        """Shutdown and close the Socket.

        :return:
        """
        if not self.socket:
            return
        try:
            if self.use_ssl:
                self.socket.unwrap()
            self.socket.shutdown(socket.SHUT_RDWR)
        except (OSError, socket.error, ValueError):
            pass

        self.socket.close()

    def _get_socket_addresses(self):
        """Get Socket address information.

        :rtype: list
        """
        family = socket.AF_UNSPEC
        if not socket.has_ipv6:
            family = socket.AF_INET
        try:
            addresses = socket.getaddrinfo(self._parameters['hostname'],
                                           self._parameters['port'], family,
                                           socket.SOCK_STREAM)
        except socket.gaierror as why:
            raise AMQPConnectionError(why)
        return addresses

    def _find_address_and_connect(self, addresses):
        """Find and connect to the appropriate address.

        :param addresses:

        :raises AMQPConnectionError: Raises if the connection
                                     encountered an error.

        :rtype: socket.socket
        """
        error_message = None
        for address in addresses:
            sock = self._create_socket(socket_family=address[0])
            try:
                sock.connect(address[4])
            except (IOError, OSError) as why:
                error_message = why.strerror
                continue
            return sock
        raise AMQPConnectionError(
            'Could not connect to %s:%d error: %s' % (
                self._parameters['hostname'], self._parameters['port'],
                error_message
            )
        )

    def _create_socket(self, socket_family):
        """Create Socket.

        :param int socket_family:
        :rtype: socket.socket
        """
        sock = socket.socket(socket_family, socket.SOCK_STREAM, 0)
        sock.settimeout(self._parameters['timeout'] or None)
        if self.use_ssl:
            if not compatibility.SSL_SUPPORTED:
                raise AMQPConnectionError(
                    'Python not compiled with support for TLSv1 or higher'
                )
            sock = self._ssl_wrap_socket(sock)
        return sock

    def _ssl_wrap_socket(self, sock):
        """Wrap SSLSocket around the Socket.

        :param socket.socket sock:
        :rtype: SSLSocket
        """
        context = self._parameters['ssl_options'].get('context')
        if context is not None:
            hostname = self._parameters['ssl_options'].get('server_hostname')
            return context.wrap_socket(
                sock, do_handshake_on_connect=True,
                server_hostname=hostname
            )
        hostname = self._parameters['hostname']
        context = ssl.SSLContext(ssl.PROTOCOL_TLS)
        mode = self._parameters['ssl_options'].get('verify_mode', 'none')
        if mode.lower() == 'required':
            context.verify_mode = ssl.CERT_REQUIRED
        else:
            context.verify_mode = ssl.CERT_NONE
        check = self._parameters['ssl_options'].get('check_hostname', False)
        context.check_hostname = check
        context.load_default_certs()
        return context.wrap_socket(sock, do_handshake_on_connect=True,
                                   server_hostname=hostname)

    def _create_inbound_thread(self):
        """Internal Thread that handles all incoming traffic.

        :rtype: threading.Thread
        """
        inbound_thread = threading.Thread(target=self._process_incoming_data,
                                          name=__name__)
        inbound_thread.daemon = True
        inbound_thread.start()
        return inbound_thread

    def _process_incoming_data(self):
        """Retrieve and process any incoming data.

        :return:
        """
        while self._running.is_set():
            if self.poller.is_ready:
                self.data_in += self._receive()
                self.data_in = self._on_read_impl(self.data_in)

    def _receive(self):
        """Receive any incoming socket data.

            If an error is thrown, handle it and return an empty string.

        :return: data_in
        :rtype: bytes
        """
        data_in = EMPTY_BUFFER
        try:
            data_in = self._read_from_socket()
            if len(data_in) == 0:
                raise socket.error("connection closed by server")
        except socket.timeout:
            pass
        except compatibility.SSLWantReadError:
            # NOTE(visobet): Retry if the non-blocking socket does not
            # have any meaningful data ready.
            pass
        except (IOError, OSError, ValueError) as why:
            if why.args[0] not in (EWOULDBLOCK, EAGAIN):
                self._exceptions.append(AMQPConnectionError(why))
                if self._running.is_set():
                    LOGGER.warning(
                        "Stopping inbound thread due to %s", why,
                        exc_info=True
                    )
                self._running.clear()
        return data_in

    def _read_from_socket(self):
        """Read data from the socket.

        :rtype: bytes
        """
        if not self.use_ssl:
            if not self.socket:
                raise socket.error('connection/socket error')
            return self.socket.recv(MAX_FRAME_SIZE)

        with self._rd_lock:
            if not self.socket:
                raise socket.error('connection/socket error')
            return self.socket.read(MAX_FRAME_SIZE)
